'''
Define utilities for building detailed landing phases and the typical landing trajectory.

Classes
-------
LandingApproachToMicP3 : a phase builder for moving from descent to the mic location P3;
this phase is required for acoustic calculations

LandingMicP3ToObstacle : a phase builder for moving from the mic location P3 to the start
of the runway, just above the required clearance height; this phase is required for
acoustic calculations

LandingObstacleToFlare : a phase builder for moving from the start of the runway, just
above the required clearance height, to the start of a maneuver to help soften the impact
of touchdown

LandingFlareToTouchdown : a phase builder for moving through a maneuver to help soften
the impact of touchdown

LandingTouchdownToNoseDown : a phase builder for rotating the nose down after touchdown

LandingNoseDownToStop : a phase builder for the final phase of landing, from nose down to
full stop

LandingTrajectory : a trajectory builder for detailed landing
'''
import dymos as dm
import openmdao.api as om

from aviary.mission.flops_based.ode.landing_ode import FlareODE, LandingODE
from aviary.mission.flops_based.phases.detailed_takeoff_phases import \
    TakeoffTrajectory as _TakeoffTrajectory
from aviary.mission.flops_based.phases.detailed_takeoff_phases import \
    _init_initial_guess_meta_data
from aviary.mission.phase_builder_base import PhaseBuilderBase
from aviary.mission.initial_guess_builders import InitialGuessControl, InitialGuessParameter, InitialGuessPolynomialControl, InitialGuessState, InitialGuessTime
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import setup_trajectory_params
from aviary.variable_info.variables import Dynamic, Mission


@_init_initial_guess_meta_data
class LandingApproachToMicP3(PhaseBuilderBase):
    '''
    Define a phase builder for moving from descent to the mic location P3. This phase is
    required for acoustic calculations.

    Attributes
    ----------
    name : str ('landing_approach')
        object label

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

        supported options:
            - max_duration (100.0, 's')
            - duration_ref (1.0, 's')
            - initial_ref (10.0, 's')
            - distance_max (1000.0, 'ft')
            - max_velocity (100.0, 'ft/s')
            - altitude_ref (1.0, 'ft')
            - lower_angle_of_attack (-10.0, 'deg')
            - upper_angle_of_attack (15.0, 'deg')
            - angle_of_attack_ref (10.0, 'deg')
            - initial_height (1.0, 'ft')

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

        supported options:
            - times
            - range
            - velocity
            - mass
            - throttle
            - angle_of_attack
            - altitude
            - Dynamic.Mission.FLIGHT_PATH_ANGLE

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = ()

    # region : derived type customization points
    _meta_data_ = {}

    default_name = 'landing_approach'

    default_ode_class = LandingODE
    # endregion : derived type customization points

    def build_phase(self, aviary_options: AviaryValues = None):
        '''
        Return a new phase object for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (emtpy)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        phase: dm.Phase = super().build_phase(aviary_options)

        user_options: AviaryValues = self.user_options

        max_duration, units = user_options.get_item('max_duration')
        duration_ref = user_options.get_val('duration_ref', units)
        initial_ref = user_options.get_val('initial_ref', units)

        phase.set_time_options(
            duration_bounds=(1, max_duration),
            duration_ref=duration_ref, initial_ref=initial_ref,
            units=units)

        distance_max, units = user_options.get_item('distance_max')

        phase.add_state(
            Dynamic.Mission.DISTANCE, fix_initial=False, fix_final=False,
            upper=0, ref=distance_max,
            defect_ref=distance_max, units=units,
            rate_source=Dynamic.Mission.DISTANCE_RATE)

        altitude_ref, units = user_options.get_item('altitude_ref')

        phase.add_state(
            Dynamic.Mission.ALTITUDE, fix_initial=False, fix_final=False,
            ref=altitude_ref,
            defect_ref=altitude_ref, units=units,
            rate_source=Dynamic.Mission.ALTITUDE_RATE)

        max_velocity, units = user_options.get_item('max_velocity')

        phase.add_state(
            Dynamic.Mission.VELOCITY, fix_initial=False, fix_final=False,
            lower=0, ref=max_velocity,
            defect_ref=max_velocity, units=units,
            rate_source=Dynamic.Mission.VELOCITY_RATE)

        phase.add_control(Dynamic.Mission.FLIGHT_PATH_ANGLE, opt=False, fix_initial=True)

        phase.add_state(
            Dynamic.Mission.MASS, fix_initial=True, fix_final=False,
            lower=0.0, ref=5e4, defect_ref=5e4, units='kg',
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            targets=Dynamic.Mission.MASS,
        )

        phase.add_control(
            Dynamic.Mission.THROTTLE,
            targets=Dynamic.Mission.THROTTLE, units='unitless',
            opt=False
        )

        units = 'deg'
        lower_angle_of_attack = user_options.get_val('lower_angle_of_attack', units)
        upper_angle_of_attack = user_options.get_val('upper_angle_of_attack', units)
        angle_of_attack_ref = user_options.get_val('angle_of_attack_ref', units)

        phase.add_control(
            'angle_of_attack', opt=True, units=units,
            upper=upper_angle_of_attack, lower=lower_angle_of_attack,
            ref=angle_of_attack_ref
        )

        phase.add_timeseries_output(
            Dynamic.Mission.DRAG, output_name=Dynamic.Mission.DRAG, units='lbf'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL,
            output_name=Dynamic.Mission.THRUST_TOTAL, units='lbf'
        )

        initial_height, units = user_options.get_item('initial_height')

        airport_altitude = aviary_options.get_val(
            Mission.Landing.AIRPORT_ALTITUDE, units)

        h = initial_height + airport_altitude

        phase.add_boundary_constraint(
            Dynamic.Mission.ALTITUDE, loc='initial', equals=h, ref=h, units=units, linear=True)

        return phase

    def make_default_transcription(self):
        '''
        Return a transcription object to be used by default in build_phase.
        '''
        transcription = dm.Radau(num_segments=5, order=3, compressed=True)

        return transcription

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        return {
            'climbing': True,
            'friction_key': Mission.Takeoff.ROLLING_FRICTION_COEFFICIENT}


LandingApproachToMicP3._add_meta_data('max_duration', val=100., units='s')

LandingApproachToMicP3._add_meta_data('duration_ref', val=1., units='s')

LandingApproachToMicP3._add_meta_data('initial_ref', val=10.0, units='s')

LandingApproachToMicP3._add_meta_data('distance_max', val=1000., units='ft')

LandingApproachToMicP3._add_meta_data('max_velocity', val=100., units='ft/s')

LandingApproachToMicP3._add_meta_data('altitude_ref', val=1., units='ft')

LandingApproachToMicP3._add_meta_data('lower_angle_of_attack', val=-10., units='deg')

LandingApproachToMicP3._add_meta_data('upper_angle_of_attack', val=15., units='deg')

LandingApproachToMicP3._add_meta_data('angle_of_attack_ref', val=10., units='deg')

LandingApproachToMicP3._add_meta_data('initial_height', val=1., units='ft')

LandingApproachToMicP3._add_initial_guess_meta_data(
    InitialGuessControl('angle_of_attack'))

LandingApproachToMicP3._add_initial_guess_meta_data(InitialGuessState('altitude'))

LandingApproachToMicP3._add_initial_guess_meta_data(
    InitialGuessControl(Dynamic.Mission.FLIGHT_PATH_ANGLE))


# @_init_initial_guess_meta_data  # <--- inherited from base class
class LandingMicP3ToObstacle(LandingApproachToMicP3):
    '''
    Define a phase builder for moving from the mic location P3 to the start of the
    runway, just above the required clearance height. This phase is required for acoustic
    calculations.

    Attributes
    ----------
    name : str ('landing_approach')
        object label

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

        supported options:
            - max_duration (100.0, 's')
            - duration_ref (1.0, 's')
            - initial_ref (10.0, 's')
            - distance_max (1000.0, 'ft')
            - max_velocity (100.0, 'ft/s')
            - altitude_ref (1.0, 'ft')
            - lower_angle_of_attack (-10.0, 'deg')
            - upper_angle_of_attack (15.0, 'deg')
            - angle_of_attack_ref (10.0, 'deg')
            - initial_height (1.0, 'ft')

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

        supported options:
            - times
            - range
            - velocity
            - mass
            - throttle
            - angle_of_attack
            - altitude
            - Dynamic.Mission.FLIGHT_PATH_ANGLE

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = ()

    # region : derived type customization points
    # _meta_data_ = {}  # <--- inherited from base class

    default_name = 'landing_mic_p3'
    # endregion : derived type customization points

    def build_phase(self, aviary_options: AviaryValues = None):
        '''
        Return a new phase object for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (emtpy)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        phase: dm.Phase = super().build_phase(aviary_options)

        # at the moment, these state options are the only differences between phases of
        # this class and phases of its base class
        phase.set_state_options(Dynamic.Mission.DISTANCE, fix_final=True)
        phase.set_state_options(Dynamic.Mission.VELOCITY, fix_final=True)
        phase.set_state_options(Dynamic.Mission.MASS, fix_initial=False)

        return phase


@_init_initial_guess_meta_data
class LandingObstacleToFlare(PhaseBuilderBase):
    '''
    Define a phase builder for moving from the start of the runway, just above the
    required clearance height, to the start of a maneuver to help soften the impact of
    touchdown.

    Attributes
    ----------
    name : str ('landing_obstacle')
        object label

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

        supported options:
            - max_duration (100.0, 's')
            - distance_max (1000.0, 'ft')
            - max_velocity (100.0, 'ft/s')
            - altitude_ref (1.0, 'ft')

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

        supported options:
            - times
            - range
            - velocity
            - mass
            - throttle
            - angle_of_attack
            - altitude
            - Dynamic.Mission.FLIGHT_PATH_ANGLE

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = ()

    # region : derived type customization points
    _meta_data_ = {}

    default_name = 'landing_obstacle'

    default_ode_class = LandingODE
    # endregion : derived type customization points

    def build_phase(self, aviary_options: AviaryValues = None):
        '''
        Return a new phase object for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (emtpy)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        phase: dm.Phase = super().build_phase(aviary_options)

        user_options: AviaryValues = self.user_options

        max_duration, units = user_options.get_item('max_duration')

        phase.set_time_options(
            fix_initial=True, duration_bounds=(1, max_duration),
            units=units)

        distance_max, units = user_options.get_item('distance_max')

        phase.add_state(
            Dynamic.Mission.DISTANCE, fix_initial=True, lower=0, ref=distance_max,
            defect_ref=distance_max, units=units,
            rate_source=Dynamic.Mission.DISTANCE_RATE)

        altitude_ref, units = user_options.get_item('altitude_ref')

        phase.add_state(
            Dynamic.Mission.ALTITUDE, fix_initial=False, lower=0, ref=altitude_ref,
            defect_ref=altitude_ref, units=units,
            rate_source=Dynamic.Mission.ALTITUDE_RATE)

        max_velocity, units = user_options.get_item('max_velocity')

        phase.add_state(
            Dynamic.Mission.VELOCITY, fix_initial=True, lower=0, ref=max_velocity,
            defect_ref=max_velocity, units=units,
            rate_source=Dynamic.Mission.VELOCITY_RATE)

        phase.add_control(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                          opt=False, fix_initial=False)

        phase.add_state(
            Dynamic.Mission.MASS, fix_initial=False, fix_final=False,
            lower=0.0, ref=5e4, defect_ref=5e4, units='kg',
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            targets=Dynamic.Mission.MASS,
        )

        phase.add_control(
            Dynamic.Mission.THROTTLE,
            targets=Dynamic.Mission.THROTTLE, units='unitless',
            opt=False
        )

        phase.add_control('angle_of_attack', opt=False, units='deg')

        phase.add_timeseries_output(
            Dynamic.Mission.DRAG, output_name=Dynamic.Mission.DRAG, units='lbf'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL,
            output_name=Dynamic.Mission.THRUST_TOTAL, units='lbf'
        )

        obstacle_height, units = aviary_options.get_item(
            Mission.Landing.OBSTACLE_HEIGHT)

        if obstacle_height is None:
            raise TypeError(
                f'missing required aviary_option: {Mission.Landing.OBSTACLE_HEIGHT}')

        airport_altitude = aviary_options.get_val(
            Mission.Landing.AIRPORT_ALTITUDE, units)

        h = obstacle_height + airport_altitude

        phase.add_boundary_constraint(
            Dynamic.Mission.ALTITUDE, loc='initial', equals=h, ref=h, units=units, linear=True)

        return phase

    def make_default_transcription(self):
        '''
        Return a transcription object to be used by default in build_phase.
        '''
        transcription = dm.Radau(num_segments=5, order=3, compressed=True)

        return transcription

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        return {
            'climbing': True,
            'friction_key': Mission.Takeoff.ROLLING_FRICTION_COEFFICIENT}


LandingObstacleToFlare._add_meta_data('max_duration', val=100., units='s')

LandingObstacleToFlare._add_meta_data('distance_max', val=1000., units='ft')

LandingObstacleToFlare._add_meta_data('max_velocity', val=100., units='ft/s')

LandingObstacleToFlare._add_meta_data('altitude_ref', val=1., units='ft')

LandingObstacleToFlare._add_initial_guess_meta_data(
    InitialGuessControl('angle_of_attack'))

LandingObstacleToFlare._add_initial_guess_meta_data(InitialGuessState('altitude'))

LandingObstacleToFlare._add_initial_guess_meta_data(
    InitialGuessControl(Dynamic.Mission.FLIGHT_PATH_ANGLE))


@_init_initial_guess_meta_data
class LandingFlareToTouchdown(PhaseBuilderBase):
    '''
    Define a phase builder for moving through a maneuver to help soften the impact of
    touchdown.

    Attributes
    ----------
    name : str ('landing_flare')
        object label

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

        supported options:
            - max_duration (100.0, 's')
            - duration_ref (1.0, 's')
            - initial_ref (10.0, 's')
            - distance_max (1000.0, 'ft')
            - max_velocity (100.0, 'ft/s')
            - altitude_ref (1.0, 'ft')
            - lower_angle_of_attack (-10.0, 'deg')
            - upper_angle_of_attack (15.0, 'deg')
            - angle_of_attack_ref (10.0, 'deg')

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

        supported options:
            - times
            - range
            - velocity
            - mass
            - throttle
            - angle_of_attack
            - altitude
            - Dynamic.Mission.FLIGHT_PATH_ANGLE

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = ()

    # region : derived type customization points
    _meta_data_ = {}

    default_name = 'landing_flare'

    default_ode_class = FlareODE
    # endregion : derived type customization points

    def build_phase(self, aviary_options: AviaryValues = None):
        '''
        Return a new phase object for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (emtpy)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        phase: dm.Phase = super().build_phase(aviary_options)

        user_options: AviaryValues = self.user_options

        max_duration, units = user_options.get_item('max_duration')
        duration_ref = user_options.get_val('duration_ref', units)
        initial_ref = user_options.get_val('initial_ref', units)

        phase.set_time_options(
            fix_initial=False, duration_bounds=(1, max_duration),
            duration_ref=duration_ref, initial_ref=initial_ref,
            units=units)

        distance_max, units = user_options.get_item('distance_max')

        phase.add_state(
            Dynamic.Mission.DISTANCE, fix_initial=False, lower=0, ref=distance_max,
            defect_ref=distance_max, units=units,
            rate_source=Dynamic.Mission.DISTANCE_RATE)

        altitude_ref, units = user_options.get_item('altitude_ref')

        phase.add_state(
            Dynamic.Mission.ALTITUDE, fix_initial=False, fix_final=True,
            lower=0, ref=altitude_ref,
            defect_ref=altitude_ref, units=units,
            rate_source=Dynamic.Mission.ALTITUDE_RATE)

        max_velocity, units = user_options.get_item('max_velocity')

        phase.add_state(
            Dynamic.Mission.VELOCITY, fix_initial=False, lower=0, ref=max_velocity,
            defect_ref=max_velocity, units=units,
            rate_source=Dynamic.Mission.VELOCITY_RATE)

        phase.add_control(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                          fix_initial=False, opt=False)

        phase.add_state(
            Dynamic.Mission.MASS, fix_initial=False, fix_final=False,
            lower=0.0, ref=5e4, defect_ref=5e4, units='kg',
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            targets=Dynamic.Mission.MASS,
        )

        # TODO: Upper limit is a bit of a hack. It hopefully won't be needed if we
        # can get some other constraints working.
        phase.add_control(
            Dynamic.Mission.THROTTLE,
            targets=Dynamic.Mission.THROTTLE, units='unitless',
            lower=0.0, upper=0.2,
            opt=True
        )

        units = 'deg'
        lower_angle_of_attack = user_options.get_val('lower_angle_of_attack', units)
        upper_angle_of_attack = user_options.get_val('upper_angle_of_attack', units)
        angle_of_attack_ref = user_options.get_val('angle_of_attack_ref', units)

        phase.add_polynomial_control(
            'angle_of_attack', opt=True, units=units, order=1,
            lower=lower_angle_of_attack, upper=upper_angle_of_attack,
            ref=angle_of_attack_ref,
            rate_targets="angle_of_attack_rate"
        )

        phase.add_timeseries_output(
            Dynamic.Mission.DRAG, output_name=Dynamic.Mission.DRAG, units='lbf'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL,
            output_name=Dynamic.Mission.THRUST_TOTAL, units='lbf'
        )

        phase.add_timeseries_output(
            'required_thrust', units='lbf'
        )

        phase.add_timeseries_output(
            'forces_perpendicular', units='lbf'
        )

        # Since the control is linear, only need to constrain one point.
        phase.add_boundary_constraint(
            'net_alpha_rate',
            equals=0.0, loc='final',
            units='deg/s'
        )

        return phase

    def make_default_transcription(self):
        '''
        Return a transcription object to be used by default in build_phase.
        '''
        transcription = dm.Radau(num_segments=5, order=3, compressed=True)

        return transcription

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        return {}


LandingFlareToTouchdown._add_meta_data('max_duration', val=100., units='s')

LandingFlareToTouchdown._add_meta_data('duration_ref', val=1., units='s')

LandingFlareToTouchdown._add_meta_data('initial_ref', val=10.0, units='s')

LandingFlareToTouchdown._add_meta_data('distance_max', val=1000., units='ft')

LandingFlareToTouchdown._add_meta_data('max_velocity', val=100., units='ft/s')

LandingFlareToTouchdown._add_meta_data('altitude_ref', val=1., units='ft')

LandingFlareToTouchdown._add_meta_data('lower_angle_of_attack', val=-10., units='deg')

LandingFlareToTouchdown._add_meta_data('upper_angle_of_attack', val=15., units='deg')

LandingFlareToTouchdown._add_meta_data('angle_of_attack_ref', val=10., units='deg')

LandingFlareToTouchdown._add_initial_guess_meta_data(
    InitialGuessPolynomialControl('angle_of_attack'))

LandingFlareToTouchdown._add_initial_guess_meta_data(InitialGuessState('altitude'))

LandingFlareToTouchdown._add_initial_guess_meta_data(
    InitialGuessControl(Dynamic.Mission.FLIGHT_PATH_ANGLE))


@_init_initial_guess_meta_data
class LandingTouchdownToNoseDown(PhaseBuilderBase):
    '''
    Define a phase builder for rotating the nose down after touchdown.

    Attributes
    ----------
    name : str ('landing_touchdown')
        object label

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

        supported options:
            - max_duration (5.0, 's')
            - duration_ref (1.0, 's')
            - initial_ref (10.0, 's')
            - distance_max (1000.0, 'ft')
            - max_velocity (100.0, 'ft/s')
            - max_angle_of_attack (10.0, 'deg')

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

        supported options:
            - times
            - range
            - velocity
            - mass
            - throttle
            - angle_of_attack

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = ()

    # region : derived type customization points
    _meta_data_ = {}

    default_name = 'landing_touchdown'

    default_ode_class = LandingODE
    # endregion : derived type customization points

    def build_phase(self, aviary_options=None):
        '''
        Return a new phase object for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (emtpy)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        phase: dm.Phase = super().build_phase(aviary_options)

        user_options: AviaryValues = self.user_options

        max_duration, units = user_options.get_item('max_duration')
        duration_ref = user_options.get_val('duration_ref', units)
        initial_ref = user_options.get_val('initial_ref', units)

        phase.set_time_options(
            fix_initial=False, duration_bounds=(1, max_duration),
            duration_ref=duration_ref, initial_ref=initial_ref,
            units=units)

        distance_max, units = user_options.get_item('distance_max')

        phase.add_state(
            Dynamic.Mission.DISTANCE, fix_initial=False, lower=0, ref=distance_max,
            defect_ref=distance_max, units=units,
            rate_source=Dynamic.Mission.DISTANCE_RATE)

        max_velocity, units = user_options.get_item('max_velocity')

        phase.add_state(
            Dynamic.Mission.VELOCITY, fix_initial=False, lower=0, ref=max_velocity,
            defect_ref=max_velocity, units=units,
            rate_source=Dynamic.Mission.VELOCITY_RATE)

        phase.add_state(
            Dynamic.Mission.MASS, fix_initial=False, fix_final=False,
            lower=0.0, ref=5e4, defect_ref=5e4, units='kg',
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            targets=Dynamic.Mission.MASS,
        )

        phase.add_control(
            Dynamic.Mission.THROTTLE,
            targets=Dynamic.Mission.THROTTLE, units='unitless',
            opt=False
        )

        units = 'deg'
        max_angle_of_attack = user_options.get_val('max_angle_of_attack', units)

        phase.add_polynomial_control(
            'angle_of_attack', opt=True, units=units, order=1,
            lower=0, upper=max_angle_of_attack, fix_final=True,
            fix_initial=False, ref=max_angle_of_attack)

        phase.add_timeseries_output(
            Dynamic.Mission.DRAG, output_name=Dynamic.Mission.DRAG, units='lbf'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL,
            output_name=Dynamic.Mission.THRUST_TOTAL, units='lbf'
        )

        return phase

    def make_default_transcription(self):
        '''
        Return a transcription object to be used by default in build_phase.
        '''
        transcription = dm.Radau(num_segments=3, order=3, compressed=True)

        return transcription

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        return {
            'climbing': False,
            'friction_key': Mission.Takeoff.ROLLING_FRICTION_COEFFICIENT}


LandingTouchdownToNoseDown._add_meta_data('max_duration', val=5.0, units='s')

LandingTouchdownToNoseDown._add_meta_data('duration_ref', val=1.0, units='s')

LandingTouchdownToNoseDown._add_meta_data('initial_ref', val=10.0, units='s')

LandingTouchdownToNoseDown._add_meta_data('distance_max', val=1000.0, units='ft')

LandingTouchdownToNoseDown._add_meta_data('max_velocity', val=100.0, units='ft/s')

LandingTouchdownToNoseDown._add_meta_data('max_angle_of_attack', val=10.0, units='deg')

LandingTouchdownToNoseDown._add_initial_guess_meta_data(
    InitialGuessPolynomialControl('angle_of_attack'))


@_init_initial_guess_meta_data
class LandingNoseDownToStop(PhaseBuilderBase):
    '''
    Define a phase builder for the final phase of landing, from nose down to full stop.

    Attributes
    ----------
    name : str ('landing_stop')
        object label

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

        supported options:
            - max_duration
            - duration_ref
            - distance_max
            - max_velocity

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

        supported options:
            - times
            - range
            - velocity
            - mass
            - throttle
            - angle_of_attack

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = ()

    # region : derived type customization points
    _meta_data_ = {}

    default_name = 'landing_stop'

    default_ode_class = LandingODE
    # endregion : derived type customization points

    def build_phase(self, aviary_options=None):
        '''
        Return a new phase object for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (emtpy)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        phase: dm.Phase = super().build_phase(aviary_options)

        user_options: AviaryValues = self.user_options

        max_duration, units = user_options.get_item('max_duration')
        duration_ref = user_options.get_val('duration_ref', units)
        initial_ref = user_options.get_val('initial_ref', units)

        phase.set_time_options(
            fix_initial=False, duration_bounds=(1, max_duration),
            duration_ref=duration_ref, initial_ref=initial_ref,
            units=units)

        distance_max, units = user_options.get_item('distance_max')

        phase.add_state(
            Dynamic.Mission.DISTANCE, fix_initial=False, fix_final=False,
            lower=0, ref=distance_max,
            defect_ref=distance_max, units=units,
            rate_source=Dynamic.Mission.DISTANCE_RATE)

        max_velocity, units = user_options.get_item('max_velocity')

        phase.add_state(
            Dynamic.Mission.VELOCITY, fix_initial=False, fix_final=True,
            lower=0, ref=max_velocity,
            defect_ref=max_velocity, units=units,
            rate_source=Dynamic.Mission.VELOCITY_RATE)

        phase.add_state(
            Dynamic.Mission.MASS, fix_initial=False, fix_final=False,
            lower=0.0, ref=5e4, defect_ref=5e4, units='kg',
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            targets=Dynamic.Mission.MASS,
        )

        # TODO: Energy phase places this under an if num_engines > 0.
        phase.add_control(
            Dynamic.Mission.THROTTLE,
            targets=Dynamic.Mission.THROTTLE, units='unitless',
            opt=False
        )

        phase.add_parameter('angle_of_attack', val=0.0, opt=False, units='deg')

        phase.add_timeseries_output(
            Dynamic.Mission.THRUST_TOTAL,
            output_name=Dynamic.Mission.THRUST_TOTAL, units='lbf'
        )

        phase.add_timeseries_output(
            Dynamic.Mission.DRAG, output_name=Dynamic.Mission.DRAG, units='lbf'
        )

        return phase

    def make_default_transcription(self):
        '''
        Return a transcription object to be used by default in build_phase.
        '''
        transcription = dm.Radau(num_segments=3, order=3, compressed=True)

        return transcription

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        return {
            'climbing': False,
            'friction_key': Mission.Takeoff.BRAKING_FRICTION_COEFFICIENT}


LandingNoseDownToStop._add_meta_data('max_duration', val=1000.0, units='s')

LandingNoseDownToStop._add_meta_data('duration_ref', val=10.0, units='s')

LandingNoseDownToStop._add_meta_data('initial_ref', val=10.0, units='s')

LandingNoseDownToStop._add_meta_data('distance_max', val=1000.0, units='ft')

LandingNoseDownToStop._add_meta_data('max_velocity', val=100.0, units='ft/s')

LandingNoseDownToStop._add_initial_guess_meta_data(
    InitialGuessParameter('angle_of_attack'))


class LandingTrajectory:
    '''
    Define a trajectory builder for detailed landing.

    Identify, collect, and call the necessary phase builders to create a typical landing
    trajectory.
    '''
    MappedPhase = _TakeoffTrajectory.MappedPhase

    default_name = 'detailed_landing'

    def __init__(self, name=None):
        if name is None:
            name = self.default_name

        self.name = name

        self._approach_to_mic_p3 = None
        self._mic_p3_to_obstacle = None
        self._obstacle_to_flare = None
        self._flare_to_touchdown = None
        self._touchdown_to_nose_down = None
        self._nose_down_to_stop = None

        self._phases = {}
        self._traj = None

    def get_phase_names(self):
        '''
        Return a list of base names for available phases.
        '''
        keys = list(self._phases)

        return keys

    def get_phase(self, key) -> dm.Phase:
        '''
        Return the phase associated with the specified base name.

        Raises
        ------
        KeyError
            if the specified base name is not found
        '''
        mapped_phase: self.MappedPhase = self._phases[key]

        return mapped_phase.phase

    def set_approach_to_mic_p3(self, phase_builder: PhaseBuilderBase):
        '''
        Assign a phase builder for moving from descent to the mic location P3. This phase
        is required for acoustic calculations
        '''
        self._approach_to_mic_p3 = phase_builder

    def set_mic_p3_to_obstacle(self, phase_builder: PhaseBuilderBase):
        '''
        Assign a phase builder for moving from the mic location P3 to the start of the
        runway, just above the required clearance height. This phase is required for
        acoustic calculations
        '''
        self._mic_p3_to_obstacle = phase_builder

    def set_obstacle_to_flare(self, phase_builder: PhaseBuilderBase):
        '''
        Assign a phase builder for moving from the start of the runway, just above the
        required clearance height, to the start of a maneuver to help soften the impact
        of touchdown.
        '''
        self._obstacle_to_flare = phase_builder

    def set_flare_to_touchdown(self, phase_builder: PhaseBuilderBase):
        '''
        Assign a phase builder for moving through a maneuver to help soften the impact of
        touchdown.
        '''
        self._flare_to_touchdown = phase_builder

    def set_touchdown_to_nose_down(self, phase_builder: PhaseBuilderBase):
        '''
        Assign a phase builder for rotating the nose down after touchdown.
        '''
        self._touchdown_to_nose_down = phase_builder

    def set_nose_down_to_stop(self, phase_builder: PhaseBuilderBase):
        '''
        Assign a phase builder for the final phase of landing, from nose down to full
        stop.
        '''
        self._nose_down_to_stop = phase_builder

    def build_trajectory(
        self, *, aviary_options: AviaryValues, model: om.Group = None,
        traj: dm.Trajectory = None
    ) -> dm.Trajectory:
        '''
        Return a new trajectory for detailed landing analysis.

        Call only after assigning phase builders for required phases.

        Parameters
        ----------
        aviary_options : AviaryValues
            collection of Aircraft/Mission specific options

        model : openmdao.api.Group (None)
            the model handling trajectory parameter setup; if `None`, trajectory
            parameter setup will not be handled

        traj : dymos.Trajectory (None)
            the trajectory to update; if `None`, a new trajetory will be updated and
            returned

        Returns
        -------
        the updated trajectory; if the specified trajectory is `None`, a new trajectory
        will be updated and returned

        Notes
        -----
        Do not modify this object or any of its referenced data between the call to
        `build_trajectory()` and the call to `apply_initial_guesses()`, or the behavior
        is undefined, no diagnostic required.
        '''
        if traj is None:
            traj = dm.Trajectory()

        self._traj = traj

        self._add_phases(aviary_options)
        self._link_phases()

        if model is not None:
            phase_names = self.get_phase_names()

            setup_trajectory_params(model, traj, aviary_options, phase_names)

        return traj

    def apply_initial_guesses(self, prob: om.Problem, traj_name):
        '''
        Call `prob.set_val()` for states/parameters/etc. for each phase in this
        trajectory.

        Call only after `build_trajectory()` and `prob.setup()`.

        Returns
        -------
        not_applied : dict[str, list[str]]
            for any phase with missing initial guesses that cannot be applied, a list of
            those missing initial guesses; if a given phase has no missing initial
            guesses, the returned mapping will not contain the name of that phase
        '''
        not_applied = {}
        phase_builder: PhaseBuilderBase = None  # type hint

        for (phase, phase_builder) in self._phases.values():
            tmp = phase_builder.apply_initial_guesses(prob, traj_name, phase)

            if tmp:
                not_applied[phase_builder.name] = tmp

        return not_applied

    def _add_phases(self, aviary_options: AviaryValues):
        self._phases = {}

        approach_to_mic_p3 = self._approach_to_mic_p3

        if approach_to_mic_p3 is not None:
            self._add_phase(approach_to_mic_p3, aviary_options)
            self._add_phase(self._mic_p3_to_obstacle, aviary_options)

        self._add_phase(self._obstacle_to_flare, aviary_options)
        self._add_phase(self._flare_to_touchdown, aviary_options)
        self._add_phase(self._touchdown_to_nose_down, aviary_options)
        self._add_phase(self._nose_down_to_stop, aviary_options)

    def _link_phases(self):
        traj: dm.Trajectory = self._traj

        basic_vars = ['time', 'distance', 'velocity', 'mass']
        ext_vars = basic_vars + ['angle_of_attack']
        full_vars = ext_vars + ['altitude']

        obstacle_name = self._obstacle_to_flare.name

        approach_to_mic_p3 = self._approach_to_mic_p3

        if approach_to_mic_p3 is not None:
            approach_p3_name = approach_to_mic_p3.name
            p3_obstacle_name = self._mic_p3_to_obstacle.name

            obstacle_vars = ['mass', 'time', 'altitude', 'angle_of_attack']
            traj.link_phases([p3_obstacle_name, obstacle_name], vars=obstacle_vars)

            pre_obs_vars = obstacle_vars + ['distance', 'velocity']
            traj.link_phases([approach_p3_name, p3_obstacle_name], vars=pre_obs_vars)

        flare_name = self._flare_to_touchdown.name

        traj.link_phases([obstacle_name, flare_name], vars=full_vars)

        touchdown_name = self._touchdown_to_nose_down.name

        traj.link_phases([flare_name, touchdown_name], vars=ext_vars)

        nose_down_name = self._nose_down_to_stop.name

        traj.link_phases([touchdown_name, nose_down_name], vars=basic_vars)

    def _add_phase(self, phase_builder: PhaseBuilderBase, aviary_options: AviaryValues):
        name = phase_builder.name
        phase = phase_builder.build_phase(aviary_options)

        self._traj.add_phase(name, phase)

        self._phases[name] = self.MappedPhase(phase, phase_builder)



import numpy as np

import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission

check = 1


def sigX(x):
    sig = 1 / (1 + np.exp(-check * x))

    return sig


def dSigXdX(x):
    # avoid overflow in squared term, underflow seems to be ok
    if x[0] < -300:
        return 0
    derivative = -1 / (1 + np.exp(-check * x)) ** 2 * (-check * np.exp(-check * x))

    return derivative


class BodyTankCalculations(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Fuel.WING_VOLUME_DESIGN, val=2)
        add_aviary_input(self, Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX, val=3)
        self.add_input(
            "fuel_mass_min",
            val=2000,
            units="lbm",
            desc="WFAMIN: minimum value of fuel mass (set when max payload is carried)",
        )
        add_aviary_input(self, Mission.Design.FUEL_MASS_REQUIRED, val=7)
        self.add_input(
            "max_wingfuel_mass", val=6, units="lbm", desc="WFWMX: maximum wingfuel mass"
        )
        add_aviary_input(self, Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, val=5000)
        add_aviary_input(self, Aircraft.Fuel.DENSITY, val=6.687, units="lbm/ft**3")
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400)
        add_aviary_input(self, Mission.Design.FUEL_MASS, val=3000)
        add_aviary_input(self, Aircraft.Design.OPERATING_MASS, val=94417)

        self.add_output(
            Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
            val=0,
            units="lbm",
            desc="WFXTRA: extra amount of fuel that is required but does not fit in wings",
        )
        self.add_output(
            "extra_fuel_volume",
            val=0,
            units="ft**3",
            desc="FVOLXTRA: excess required design fuel volume (including fuel margin) greater than geometric fuel volume of wings",
        )
        self.add_output(
            "max_extra_fuel_mass",
            val=0,
            units="lbm",
            desc="WFXTRAMX: mass of fuel that fits in extra_fuel_volume",
        )
        self.add_output(
            "wingfuel_mass_min",
            val=0,
            units="lbm",
            desc="WFWMIN: minimum wing fuel mass",
        )
        add_aviary_output(self, Aircraft.Fuel.TOTAL_CAPACITY, val=0)

        self.declare_partials(Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY, [
                              Mission.Design.FUEL_MASS_REQUIRED, "max_wingfuel_mass"])
        self.declare_partials(
            "extra_fuel_volume",
            [Aircraft.Fuel.WING_VOLUME_DESIGN, Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
                Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX],
        )
        self.declare_partials(
            "max_extra_fuel_mass",
            [
                Aircraft.Fuel.WING_VOLUME_DESIGN,
                Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
                Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX,
                Aircraft.Fuel.DENSITY,
            ],
        )
        self.declare_partials(
            "wingfuel_mass_min",
            [
                "fuel_mass_min",
                Aircraft.Fuel.WING_VOLUME_DESIGN,
                Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
                Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX,
                Aircraft.Fuel.DENSITY,
            ],
        )
        self.declare_partials(
            Aircraft.Fuel.TOTAL_CAPACITY,
            [
                Mission.Design.FUEL_MASS,
                Mission.Design.FUEL_MASS_REQUIRED,
                "max_wingfuel_mass",
                Mission.Design.GROSS_MASS,
                Aircraft.Design.OPERATING_MASS,
            ],
        )

    def compute(self, inputs, outputs):

        design_fuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_DESIGN]
        max_wingfuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX]
        fuel_wt_min = inputs["fuel_mass_min"] * GRAV_ENGLISH_LBM
        req_fuel_wt = inputs[Mission.Design.FUEL_MASS_REQUIRED] * \
            GRAV_ENGLISH_LBM
        max_wingfuel_wt = inputs["max_wingfuel_mass"] * GRAV_ENGLISH_LBM
        geometric_fuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX]
        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        fuel_wt_des = inputs[Mission.Design.FUEL_MASS] * GRAV_ENGLISH_LBM
        OEW = inputs[Aircraft.Design.OPERATING_MASS] * GRAV_ENGLISH_LBM

        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')

        extra_fuel_volume = sigX(design_fuel_vol - max_wingfuel_vol) * (
            design_fuel_vol - geometric_fuel_vol
        )

        if (req_fuel_wt > max_wingfuel_wt) and (design_fuel_vol > max_wingfuel_vol):
            print("Warning: req_fuel_mass > max_wingfuel_mass, adding a body tank")
        if (req_fuel_wt < max_wingfuel_wt) and (design_fuel_vol > max_wingfuel_vol):
            print("Warning: design_fuel_vol > max_wingfuel_vol, adding a body tank")

        extra_fuel_wt = req_fuel_wt - max_wingfuel_wt
        if smooth:
            extra_fuel_wt = extra_fuel_wt * sigX(extra_fuel_wt / 50)
        else:
            if extra_fuel_wt < 0:
                extra_fuel_wt = 0
        max_extra_fuel_wt = extra_fuel_volume * rho_fuel
        wingfuel_wt_min = fuel_wt_min - max_extra_fuel_wt
        wingfuel_wt_min = wingfuel_wt_min * sigX(wingfuel_wt_min)
        max_fuel_avail_est = fuel_wt_des + extra_fuel_wt
        max_fuel_avail_new = gross_wt_initial - OEW
        est_GTOW = OEW + max_fuel_avail_est
        max_fuel_avail = max_fuel_avail_est * sigX(
            (gross_wt_initial - est_GTOW) / 110
        ) + max_fuel_avail_new * sigX((est_GTOW - gross_wt_initial) / 110)

        outputs[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY] = \
            extra_fuel_wt / GRAV_ENGLISH_LBM
        outputs["extra_fuel_volume"] = extra_fuel_volume
        outputs["max_extra_fuel_mass"] = max_extra_fuel_wt / GRAV_ENGLISH_LBM
        outputs["wingfuel_mass_min"] = wingfuel_wt_min / GRAV_ENGLISH_LBM
        outputs[Aircraft.Fuel.TOTAL_CAPACITY] = \
            max_fuel_avail / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):

        design_fuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_DESIGN]
        max_wingfuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX]
        fuel_wt_min = inputs["fuel_mass_min"] * GRAV_ENGLISH_LBM
        req_fuel_wt = inputs[Mission.Design.FUEL_MASS_REQUIRED] * \
            GRAV_ENGLISH_LBM
        max_wingfuel_wt = inputs["max_wingfuel_mass"] * GRAV_ENGLISH_LBM
        geometric_fuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX]
        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        fuel_wt_des = inputs[Mission.Design.FUEL_MASS] * GRAV_ENGLISH_LBM
        OEW = inputs[Aircraft.Design.OPERATING_MASS] * GRAV_ENGLISH_LBM

        smooth = self.options["aviary_options"].get_val(
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES, units='unitless')

        extra_fuel_volume = sigX(design_fuel_vol - max_wingfuel_vol) * (
            design_fuel_vol - geometric_fuel_vol
        )
        max_extra_fuel_wt = extra_fuel_volume * rho_fuel

        extra_fuel_wt = req_fuel_wt - max_wingfuel_wt
        if smooth:
            extra_fuel_wt = extra_fuel_wt * sigX(extra_fuel_wt / 50)
        else:
            if extra_fuel_wt < 0:
                extra_fuel_wt = 0
        max_fuel_avail_est = fuel_wt_des + extra_fuel_wt
        max_fuel_avail_new = gross_wt_initial - OEW
        est_GTOW = OEW + max_fuel_avail_est

        if smooth:
            dExtraFuelWt_dReqFuelWt = (
                sigX((req_fuel_wt - max_wingfuel_wt) / 50)
                + (req_fuel_wt - max_wingfuel_wt)
                * dSigXdX((req_fuel_wt - max_wingfuel_wt) / 50)
                * 1
                / 50
            )
            dExtraFuelWt_dMaxWingfuelWt = -1 * sigX(
                (req_fuel_wt - max_wingfuel_wt) / 50
            ) + (req_fuel_wt - max_wingfuel_wt) * dSigXdX(
                (req_fuel_wt - max_wingfuel_wt) / 50
            ) * (
                -1 / 50
            )
        else:
            dExtraFuelWt_dReqFuelWt = 1 if extra_fuel_wt > 0 else 0
            dExtraFuelWt_dMaxWingfuelWt = -1 if extra_fuel_wt > 0 else 0
        J[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
            Mission.Design.FUEL_MASS_REQUIRED] = dExtraFuelWt_dReqFuelWt
        J[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY,
            "max_wingfuel_mass"] = dExtraFuelWt_dMaxWingfuelWt

        J["extra_fuel_volume", Aircraft.Fuel.WING_VOLUME_DESIGN] = dSigXdX(
            design_fuel_vol - max_wingfuel_vol
        ) * (design_fuel_vol - geometric_fuel_vol) + sigX(
            design_fuel_vol - max_wingfuel_vol
        )
        J["extra_fuel_volume", Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX] = (
            dSigXdX(design_fuel_vol - max_wingfuel_vol)
            * (-1)
            * (design_fuel_vol - geometric_fuel_vol)
        )
        J["extra_fuel_volume", Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = -sigX(
            design_fuel_vol - max_wingfuel_vol
        )

        dMEFWt_dDesFuelVol = (
            dSigXdX(design_fuel_vol - max_wingfuel_vol)
            * (design_fuel_vol - geometric_fuel_vol)
            + sigX(design_fuel_vol - max_wingfuel_vol)
        ) * rho_fuel
        dMEFWt_dMaxWingfuelVol = (
            dSigXdX(design_fuel_vol - max_wingfuel_vol)
            * (-1)
            * (design_fuel_vol - geometric_fuel_vol)
            * rho_fuel
        )
        dMEFWt_dGeomFuelVol = (
            -sigX(design_fuel_vol - max_wingfuel_vol) * rho_fuel
        )

        J["max_extra_fuel_mass", Aircraft.Fuel.WING_VOLUME_DESIGN] = \
            dMEFWt_dDesFuelVol / GRAV_ENGLISH_LBM
        J["max_extra_fuel_mass", Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX] = \
            dMEFWt_dMaxWingfuelVol / GRAV_ENGLISH_LBM
        J["max_extra_fuel_mass", Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = \
            dMEFWt_dGeomFuelVol / GRAV_ENGLISH_LBM
        J["max_extra_fuel_mass", Aircraft.Fuel.DENSITY] = dMEFWt_dRhoFuel = extra_fuel_volume

        J["wingfuel_mass_min", "fuel_mass_min"] = sigX(fuel_wt_min - max_extra_fuel_wt) + (
            fuel_wt_min - max_extra_fuel_wt
        ) * dSigXdX(fuel_wt_min - max_extra_fuel_wt)
        J["wingfuel_mass_min", Aircraft.Fuel.WING_VOLUME_DESIGN] = ((
            fuel_wt_min - max_extra_fuel_wt
        ) * dSigXdX(fuel_wt_min - max_extra_fuel_wt) * (
            -dMEFWt_dDesFuelVol
        ) - dMEFWt_dDesFuelVol * sigX(
            fuel_wt_min - max_extra_fuel_wt
        )) / GRAV_ENGLISH_LBM
        J["wingfuel_mass_min", Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX] = (fuel_wt_min * dSigXdX(
            fuel_wt_min - max_extra_fuel_wt
        ) * (-dMEFWt_dMaxWingfuelVol) - (
            dMEFWt_dMaxWingfuelVol * sigX(fuel_wt_min - max_extra_fuel_wt)
            + max_extra_fuel_wt
            * dSigXdX(fuel_wt_min - max_extra_fuel_wt)
            * (-dMEFWt_dMaxWingfuelVol)
        )) / GRAV_ENGLISH_LBM
        J["wingfuel_mass_min", Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = (fuel_wt_min * dSigXdX(
            fuel_wt_min - max_extra_fuel_wt
        ) * (-dMEFWt_dGeomFuelVol) - (
            dMEFWt_dGeomFuelVol * sigX(fuel_wt_min - max_extra_fuel_wt)
            + max_extra_fuel_wt
            * dSigXdX(fuel_wt_min - max_extra_fuel_wt)
            * (-dMEFWt_dGeomFuelVol)
        )) / GRAV_ENGLISH_LBM
        J["wingfuel_mass_min", Aircraft.Fuel.DENSITY] = fuel_wt_min * dSigXdX(
            fuel_wt_min - max_extra_fuel_wt
        ) * (-dMEFWt_dRhoFuel) - (
            dMEFWt_dRhoFuel * sigX(fuel_wt_min - max_extra_fuel_wt)
            + max_extra_fuel_wt
            * dSigXdX(fuel_wt_min - max_extra_fuel_wt)
            * (-dMEFWt_dRhoFuel)
        )

        int1 = sigX((gross_wt_initial - est_GTOW) / 110)
        int2 = sigX((est_GTOW - gross_wt_initial) / 110)
        max_fuel_avail = max_fuel_avail_est * int1 + max_fuel_avail_new * int2

        dInt1_dFuelWtDes = dSigXdX((gross_wt_initial - est_GTOW) / 110) * (-1 / 110)
        dInt1_dReqFuelWt = dSigXdX((gross_wt_initial - est_GTOW) / 110) * (
            -dExtraFuelWt_dReqFuelWt
        )
        dInt1_dMaxWingfuelWt = dSigXdX((gross_wt_initial - est_GTOW) / 110) * (
            -dExtraFuelWt_dMaxWingfuelWt
        )

        dInt2_dOEM = dSigXdX((est_GTOW - gross_wt_initial) / 110) / 110
        dInt2_dFuelWtDes = dSigXdX((est_GTOW - gross_wt_initial) / 110) / 110
        dInt2_dReqFuelWt = (
            dSigXdX((est_GTOW - gross_wt_initial) / 110) * dExtraFuelWt_dReqFuelWt
        )
        dInt2_dMaxWingfuelWt = (
            dSigXdX((est_GTOW - gross_wt_initial) / 110) * dExtraFuelWt_dMaxWingfuelWt
        )
        dInt2_dGTOW = dSigXdX((est_GTOW - gross_wt_initial) / 110) * (-1 / 110)

        dMFAest_dFuelWtDes = 1
        dMFAest_dReqFuelWt = dExtraFuelWt_dReqFuelWt
        dMFAest_dMaxWingfuelWt = dExtraFuelWt_dMaxWingfuelWt

        J[Aircraft.Fuel.TOTAL_CAPACITY, Mission.Design.FUEL_MASS] = (
            int1
            + max_fuel_avail_est * dInt1_dFuelWtDes
            + max_fuel_avail_new * dInt2_dFuelWtDes
        )
        J[Aircraft.Fuel.TOTAL_CAPACITY, Mission.Design.FUEL_MASS_REQUIRED] = (
            dMFAest_dReqFuelWt * int1
            + max_fuel_avail_est * dInt1_dReqFuelWt
            + max_fuel_avail_new * dInt2_dReqFuelWt
        )
        J[Aircraft.Fuel.TOTAL_CAPACITY, "max_wingfuel_mass"] = (
            dMFAest_dMaxWingfuelWt * int1
            + max_fuel_avail_est * dInt1_dMaxWingfuelWt
            + max_fuel_avail_new * dInt2_dMaxWingfuelWt
        )
        J[Aircraft.Fuel.TOTAL_CAPACITY, Mission.Design.GROSS_MASS] = (
            max_fuel_avail_est * dInt2_dGTOW
            + 1 * int2
            + max_fuel_avail_new * dInt2_dGTOW
        )
        J[Aircraft.Fuel.TOTAL_CAPACITY, Aircraft.Design.OPERATING_MASS] = (
            max_fuel_avail_est * dInt2_dOEM - 1 * int2 + max_fuel_avail_new * dInt2_dOEM
        )


class FuelAndOEMOutputs(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Fuel.DENSITY, val=6.687, units="lbm/ft**3")
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400)
        add_aviary_input(self, Aircraft.Propulsion.MASS, val=16129)
        add_aviary_input(self, Aircraft.Controls.TOTAL_MASS, val=3895.0)
        add_aviary_input(self, Aircraft.Design.STRUCTURE_MASS, val=50461.0)
        add_aviary_input(self, Aircraft.Design.FIXED_EQUIPMENT_MASS, val=21089.0)
        add_aviary_input(self, Aircraft.Design.FIXED_USEFUL_LOAD, val=4932.0)
        add_aviary_input(self, Mission.Design.FUEL_MASS_REQUIRED, val=42892.0)
        add_aviary_input(self, Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, val=5000)
        add_aviary_input(self, Aircraft.Fuel.FUEL_MARGIN, val=10)
        add_aviary_input(self, Aircraft.Fuel.TOTAL_CAPACITY, val=46000)

        self.add_output(
            "OEM_wingfuel_mass",
            val=0,
            units="lbm",
            desc="WFWOWE: wing fuel mass when operating empty",
        )
        self.add_output(
            "OEM_fuel_vol",
            val=0,
            units="ft**3",
            desc="FVOLW: wing tank fuel volume when carrying maximum fuel",
        )
        add_aviary_output(self, Aircraft.Fuel.WING_VOLUME_DESIGN, val=0)
        add_aviary_output(self, Aircraft.Design.OPERATING_MASS, val=0)

        self.add_output(
            "payload_mass_max_fuel",
            val=0,
            units="lbm",
            desc="WPLMXF: allowable payload mass with maximum fuel",
        )
        self.add_output(
            "volume_wingfuel_mass",
            val=0,
            units="lbm",
            desc=" mass of wing fuel based on volume, sometimes set as WFWMX in GASP, depending on if it exceeds the OEM fuel mass",
        )
        self.add_output(
            "max_wingfuel_mass", val=0, units="lbm", desc="WFWMX: maximum wingfuel mass"
        )
        add_aviary_output(self, Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX, val=0)

        self.declare_partials(
            "OEM_wingfuel_mass",
            [
                Mission.Design.GROSS_MASS,
                Aircraft.Propulsion.MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
            ],
        )
        self.declare_partials(
            "OEM_fuel_vol",
            [
                Mission.Design.GROSS_MASS,
                Aircraft.Propulsion.MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
                Aircraft.Fuel.DENSITY,
            ],
        )
        self.declare_partials(
            Aircraft.Fuel.WING_VOLUME_DESIGN, [Mission.Design.FUEL_MASS_REQUIRED,
                                               Aircraft.Fuel.DENSITY, Aircraft.Fuel.FUEL_MARGIN]
        )
        self.declare_partials(
            Aircraft.Design.OPERATING_MASS,
            [Aircraft.Propulsion.MASS, Aircraft.Controls.TOTAL_MASS, Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS, Aircraft.Design.FIXED_USEFUL_LOAD],
            val=1,
        )
        self.declare_partials(
            "payload_mass_max_fuel", [Mission.Design.GROSS_MASS], val=1
        )
        self.declare_partials(
            "payload_mass_max_fuel",
            [
                Aircraft.Fuel.TOTAL_CAPACITY,
                Aircraft.Propulsion.MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
            ],
            val=-1,
        )
        self.declare_partials(
            "volume_wingfuel_mass", [
                Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Fuel.DENSITY]
        )
        self.declare_partials(
            "max_wingfuel_mass",
            [
                Mission.Design.GROSS_MASS,
                Aircraft.Propulsion.MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
                Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX,
                Aircraft.Fuel.DENSITY,
            ],
        )
        self.declare_partials(
            Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
            [
                Mission.Design.GROSS_MASS,
                Aircraft.Propulsion.MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
                Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX,
                Aircraft.Fuel.DENSITY,
            ],
        )

    def compute(self, inputs, outputs):

        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        propulsion_wt = inputs[Aircraft.Propulsion.MASS] * GRAV_ENGLISH_LBM
        control_wt = inputs[Aircraft.Controls.TOTAL_MASS] * GRAV_ENGLISH_LBM
        struct_wt = inputs[Aircraft.Design.STRUCTURE_MASS] * GRAV_ENGLISH_LBM
        fixed_equip_wt = inputs[Aircraft.Design.FIXED_EQUIPMENT_MASS] * \
            GRAV_ENGLISH_LBM
        useful_wt = inputs[Aircraft.Design.FIXED_USEFUL_LOAD] * \
            GRAV_ENGLISH_LBM
        req_fuel_wt = inputs[Mission.Design.FUEL_MASS_REQUIRED] * \
            GRAV_ENGLISH_LBM
        geometric_fuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX]
        fuel_margin = inputs[Aircraft.Fuel.FUEL_MARGIN]
        max_fuel_avail = inputs[Aircraft.Fuel.TOTAL_CAPACITY] * \
            GRAV_ENGLISH_LBM

        OEM_wingfuel_wt = (
            gross_wt_initial
            - propulsion_wt
            - control_wt
            - struct_wt
            - fixed_equip_wt
            - useful_wt
        )

        OEM_fuel_vol = OEM_wingfuel_wt / rho_fuel
        design_fuel_vol = (1.0 + fuel_margin / 100.0) * req_fuel_wt / rho_fuel

        OEW = propulsion_wt + control_wt + struct_wt + fixed_equip_wt + useful_wt

        volume_wingfuel_wt = geometric_fuel_vol * rho_fuel
        max_wingfuel_wt = OEM_wingfuel_wt * sigX(
            (volume_wingfuel_wt - OEM_wingfuel_wt) / 95
        ) + volume_wingfuel_wt * sigX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
        payload_wt_max_fuel = gross_wt_initial - OEW - max_fuel_avail
        max_wingfuel_vol = max_wingfuel_wt / (rho_fuel)

        outputs["OEM_wingfuel_mass"] = OEM_wingfuel_wt / GRAV_ENGLISH_LBM
        outputs["OEM_fuel_vol"] = OEM_fuel_vol
        outputs[Aircraft.Fuel.WING_VOLUME_DESIGN] = design_fuel_vol
        outputs[Aircraft.Design.OPERATING_MASS] = OEW / GRAV_ENGLISH_LBM
        outputs["payload_mass_max_fuel"] = \
            payload_wt_max_fuel / GRAV_ENGLISH_LBM
        outputs["volume_wingfuel_mass"] = volume_wingfuel_wt / GRAV_ENGLISH_LBM
        outputs["max_wingfuel_mass"] = max_wingfuel_wt / GRAV_ENGLISH_LBM
        outputs[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX] = max_wingfuel_vol

    def compute_partials(self, inputs, J):

        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        propulsion_wt = inputs[Aircraft.Propulsion.MASS] * GRAV_ENGLISH_LBM
        control_wt = inputs[Aircraft.Controls.TOTAL_MASS] * GRAV_ENGLISH_LBM
        struct_wt = inputs[Aircraft.Design.STRUCTURE_MASS] * GRAV_ENGLISH_LBM
        fixed_equip_wt = inputs[Aircraft.Design.FIXED_EQUIPMENT_MASS] * \
            GRAV_ENGLISH_LBM
        useful_wt = inputs[Aircraft.Design.FIXED_USEFUL_LOAD] * \
            GRAV_ENGLISH_LBM
        req_fuel_wt = inputs[Mission.Design.FUEL_MASS_REQUIRED] * \
            GRAV_ENGLISH_LBM
        geometric_fuel_vol = inputs[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX]
        fuel_margin = inputs[Aircraft.Fuel.FUEL_MARGIN]
        max_fuel_avail = inputs[Aircraft.Fuel.TOTAL_CAPACITY] * \
            GRAV_ENGLISH_LBM

        OEM_wingfuel_wt = (
            gross_wt_initial
            - propulsion_wt
            - control_wt
            - struct_wt
            - fixed_equip_wt
            - useful_wt
        )
        volume_wingfuel_wt = geometric_fuel_vol * rho_fuel
        max_wingfuel_wt = OEM_wingfuel_wt * sigX(
            (volume_wingfuel_wt - OEM_wingfuel_wt) / 95
        ) + volume_wingfuel_wt * sigX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)

        J["OEM_wingfuel_mass", Mission.Design.GROSS_MASS] = dOEMwingfuelWt_dGTOW = 1
        J["OEM_wingfuel_mass", Aircraft.Propulsion.MASS] = dOEMwingfuelWt_dPropWt = -1
        J["OEM_wingfuel_mass", Aircraft.Controls.TOTAL_MASS] = dOEMwingfuelWt_dControlWt = -1
        J["OEM_wingfuel_mass", Aircraft.Design.STRUCTURE_MASS] = dOEMwingfuelWt_dStructWt = -1
        J["OEM_wingfuel_mass", Aircraft.Design.FIXED_EQUIPMENT_MASS] = dOEMwingfuelWt_dFEqWt = -1
        J["OEM_wingfuel_mass", Aircraft.Design.FIXED_USEFUL_LOAD] = dOEMwingfuelWt_dUsefulWt = -1

        J["OEM_fuel_vol", Mission.Design.GROSS_MASS] = \
            1 / rho_fuel * GRAV_ENGLISH_LBM
        J["OEM_fuel_vol", Aircraft.Propulsion.MASS] = \
            -1 / rho_fuel * GRAV_ENGLISH_LBM
        J["OEM_fuel_vol", Aircraft.Controls.TOTAL_MASS] = \
            -1 / rho_fuel * GRAV_ENGLISH_LBM
        J["OEM_fuel_vol", Aircraft.Design.STRUCTURE_MASS] = \
            -1 / rho_fuel * GRAV_ENGLISH_LBM
        J["OEM_fuel_vol", Aircraft.Design.FIXED_EQUIPMENT_MASS] = \
            -1 / rho_fuel * GRAV_ENGLISH_LBM
        J["OEM_fuel_vol", Aircraft.Design.FIXED_USEFUL_LOAD] = \
            -1 / rho_fuel * GRAV_ENGLISH_LBM
        J["OEM_fuel_vol", Aircraft.Fuel.DENSITY] = \
            -OEM_wingfuel_wt / rho_fuel**2 * GRAV_ENGLISH_LBM

        J[Aircraft.Fuel.WING_VOLUME_DESIGN, Mission.Design.FUEL_MASS_REQUIRED] = (
            1.0 + fuel_margin / 100.0) / rho_fuel * GRAV_ENGLISH_LBM
        J[Aircraft.Fuel.WING_VOLUME_DESIGN, Aircraft.Fuel.DENSITY] = (
            -(1.0 + fuel_margin / 100.0) * req_fuel_wt /
            rho_fuel**2 * GRAV_ENGLISH_LBM
        )
        J[Aircraft.Fuel.WING_VOLUME_DESIGN, Aircraft.Fuel.FUEL_MARGIN] = (
            1 / 100.0 * req_fuel_wt / rho_fuel
        )

        J["volume_wingfuel_mass", Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = rho_fuel / \
            GRAV_ENGLISH_LBM
        J["volume_wingfuel_mass", Aircraft.Fuel.DENSITY] = geometric_fuel_vol

        dMaxWFWt_dGTOW = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dGTOW
            + dOEMwingfuelWt_dGTOW *
            sigX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dGTOW
        )
        dMaxWFWt_dPropWt = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dPropWt
            + dOEMwingfuelWt_dPropWt *
            sigX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dPropWt
        )
        dMaxWFWt_dControlWt = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dControlWt
            + dOEMwingfuelWt_dControlWt
            * sigX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dControlWt
        )
        dMaxWFWt_dControlWt = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dControlWt
            + dOEMwingfuelWt_dControlWt
            * sigX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dControlWt
        )
        dMaxWFWt_dStructWt = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dStructWt
            + dOEMwingfuelWt_dStructWt
            * sigX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dStructWt
        )
        dMaxWFWt_dFEqWt = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dFEqWt
            + dOEMwingfuelWt_dFEqWt *
            sigX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dFEqWt
        )
        dMaxWFWt_dUsefulWt = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dUsefulWt
            + dOEMwingfuelWt_dUsefulWt
            * sigX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * dOEMwingfuelWt_dUsefulWt
        )
        dMaxWFWt_dGeomFuelVol = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * rho_fuel
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * rho_fuel
            + rho_fuel * sigX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
        )
        dMaxWFWt_dRhoFuel = (
            OEM_wingfuel_wt
            * dSigXdX((volume_wingfuel_wt - OEM_wingfuel_wt) / 95)
            * geometric_fuel_vol
            + volume_wingfuel_wt
            * dSigXdX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
            * geometric_fuel_vol
            + geometric_fuel_vol * sigX((OEM_wingfuel_wt - volume_wingfuel_wt) / 95)
        )

        J["max_wingfuel_mass", Mission.Design.GROSS_MASS] = dMaxWFWt_dGTOW
        J["max_wingfuel_mass", Aircraft.Propulsion.MASS] = dMaxWFWt_dPropWt
        J["max_wingfuel_mass", Aircraft.Controls.TOTAL_MASS] = dMaxWFWt_dControlWt
        J["max_wingfuel_mass", Aircraft.Design.STRUCTURE_MASS] = dMaxWFWt_dStructWt
        J["max_wingfuel_mass", Aircraft.Design.FIXED_EQUIPMENT_MASS] = dMaxWFWt_dFEqWt
        J["max_wingfuel_mass", Aircraft.Design.FIXED_USEFUL_LOAD] = dMaxWFWt_dUsefulWt
        J["max_wingfuel_mass", Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = dMaxWFWt_dGeomFuelVol / \
            GRAV_ENGLISH_LBM
        J["max_wingfuel_mass", Aircraft.Fuel.DENSITY] = dMaxWFWt_dRhoFuel

        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
            Mission.Design.GROSS_MASS] = dMaxWFWt_dGTOW / (rho_fuel)
        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
            Aircraft.Propulsion.MASS] = dMaxWFWt_dPropWt / (rho_fuel)
        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
            Aircraft.Controls.TOTAL_MASS] = dMaxWFWt_dControlWt / (rho_fuel)
        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
            Aircraft.Design.STRUCTURE_MASS] = dMaxWFWt_dStructWt / (rho_fuel)
        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
            Aircraft.Design.FIXED_EQUIPMENT_MASS] = dMaxWFWt_dFEqWt / (rho_fuel)
        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX,
            Aircraft.Design.FIXED_USEFUL_LOAD] = dMaxWFWt_dUsefulWt / (rho_fuel)
        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX, Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = dMaxWFWt_dGeomFuelVol / (
            rho_fuel
        )
        J[Aircraft.Fuel.WING_VOLUME_STRUCTURAL_MAX, Aircraft.Fuel.DENSITY] = dMaxWFWt_dRhoFuel / (
            rho_fuel
        ) - max_wingfuel_wt / (rho_fuel**2)


class FuelSysAndFullFuselageMass(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400)

        add_aviary_input(self, Aircraft.Wing.MASS, val=15830.0)
        self.add_input(
            "wing_mounted_mass",
            val=24446.343040697346,
            units="lbm",
            desc="WM: mass of gear and engine (everything on wing that isn`t wing itself or fuel",
        )
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041)
        add_aviary_input(self, Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        add_aviary_input(self, Mission.Design.FUEL_MASS, val=3000)
        add_aviary_input(self, Aircraft.Fuel.FUEL_MARGIN, val=10)
        self.add_input(
            "wingfuel_mass_min",
            val=32850,
            units="lbm",
            desc="WFWMIN: minimum wing fuel mass",
        )

        self.add_output(
            "fus_mass_full",
            val=0,
            units="lbm",
            desc="WX: nass of fuselage and contents, including empennage",
        )
        add_aviary_output(self, Aircraft.Fuel.FUEL_SYSTEM_MASS, val=0)

        self.declare_partials(
            Aircraft.Fuel.FUEL_SYSTEM_MASS,
            [
                Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER,
                Aircraft.Fuel.DENSITY,
                Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT,
                Mission.Design.FUEL_MASS,
                Aircraft.Fuel.FUEL_MARGIN,
            ],
        )
        self.declare_partials(
            "fus_mass_full",
            [
                Mission.Design.GROSS_MASS,
                Aircraft.Wing.MASS,
                "wingfuel_mass_min",
                "wing_mounted_mass",
            ],
        )

    def compute(self, inputs, outputs):

        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        total_wing_wt = inputs[Aircraft.Wing.MASS] * GRAV_ENGLISH_LBM
        wing_mounted_wt = inputs["wing_mounted_mass"] * GRAV_ENGLISH_LBM
        CK21 = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]
        c_mass_trend_fuel_sys = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT]
        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        fuel_wt_des = inputs[Mission.Design.FUEL_MASS] * GRAV_ENGLISH_LBM
        fuel_margin = inputs[Aircraft.Fuel.FUEL_MARGIN]
        wingfuel_wt_min = inputs["wingfuel_mass_min"] * GRAV_ENGLISH_LBM

        outputs[Aircraft.Fuel.FUEL_SYSTEM_MASS] = (
            CK21
            * (6.687 / rho_fuel)
            * c_mass_trend_fuel_sys
            * fuel_wt_des
            * (1.0 + fuel_margin / 100.0)
            / GRAV_ENGLISH_LBM
        )
        outputs["fus_mass_full"] = (
            gross_wt_initial - total_wing_wt - wingfuel_wt_min - wing_mounted_wt
        ) / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):

        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        total_wing_wt = inputs[Aircraft.Wing.MASS] * GRAV_ENGLISH_LBM
        wing_mounted_wt = inputs["wing_mounted_mass"] * GRAV_ENGLISH_LBM
        CK21 = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]
        c_mass_trend_fuel_sys = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT]
        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        fuel_wt_des = inputs[Mission.Design.FUEL_MASS] * GRAV_ENGLISH_LBM
        fuel_margin = inputs[Aircraft.Fuel.FUEL_MARGIN]
        wingfuel_wt_min = inputs["wingfuel_mass_min"] * GRAV_ENGLISH_LBM

        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER] = (
            (6.687 / rho_fuel)
            * c_mass_trend_fuel_sys
            * fuel_wt_des
            * (1.0 + fuel_margin / 100.0)
            / GRAV_ENGLISH_LBM
        )
        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.DENSITY] = (
            -CK21
            * (6.687 / rho_fuel**2)
            * c_mass_trend_fuel_sys
            * fuel_wt_des
            * (1.0 + fuel_margin / 100.0)
        )
        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT] = (
            CK21 * (6.687 / rho_fuel) * fuel_wt_des *
            (1.0 + fuel_margin / 100.0) / GRAV_ENGLISH_LBM
        )
        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Mission.Design.FUEL_MASS] = (
            CK21
            * (6.687 / rho_fuel)
            * c_mass_trend_fuel_sys
            * (1.0 + fuel_margin / 100.0)
        )
        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.FUEL_MARGIN] = (
            CK21 * (6.687 / rho_fuel) * c_mass_trend_fuel_sys *
            fuel_wt_des * (1 / 100.0) / GRAV_ENGLISH_LBM
        )

        J["fus_mass_full", Mission.Design.GROSS_MASS] = 1
        J["fus_mass_full", Aircraft.Wing.MASS] = -1
        J["fus_mass_full", "wingfuel_mass_min"] = -1
        J["fus_mass_full", "wing_mounted_mass"] = -1


class FuselageAndStructMass(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        self.add_input(
            "fus_mass_full",
            val=4000,
            units="lbm",
            desc="WX: mass of fuselage and contents, including empennage",
        )
        add_aviary_input(self, Aircraft.Fuselage.MASS_COEFFICIENT, val=128)
        add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA, val=4000)
        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=13.1)
        add_aviary_input(self, Aircraft.TailBoom.LENGTH, val=129.4)
        self.add_input(
            "pylon_len",
            val=0,
            units="ft",
            desc="ELRW: length of pylon for fuselage mounted engines",
        )
        self.add_input(
            "min_dive_vel", val=419.75918333, units="kn", desc="VDMIN: dive velocity"
        )
        add_aviary_input(self, Aircraft.Fuselage.PRESSURE_DIFFERENTIAL, val=7.5)
        add_aviary_input(self, Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893)
        self.add_input(
            "MAT",
            val=0,
            units="lbm",
            desc="WAT: Weight of the Fuselage Acoustic Treatment",
        )
        add_aviary_input(self, Aircraft.Wing.MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.Wing.MASS, val=5000)
        add_aviary_input(self, Aircraft.HorizontalTail.MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.HorizontalTail.MASS, val=2275)
        add_aviary_input(self, Aircraft.VerticalTail.MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.VerticalTail.MASS, val=2297)
        add_aviary_input(self, Aircraft.Fuselage.MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.LandingGear.TOTAL_MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.LandingGear.TOTAL_MASS, val=7511)
        add_aviary_input(self, Aircraft.Engine.POD_MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS, val=3785)
        add_aviary_input(self, Aircraft.Design.STRUCTURAL_MASS_INCREMENT, val=0)

        add_aviary_output(self, Aircraft.Design.STRUCTURE_MASS, val=0)
        add_aviary_output(self, Aircraft.Fuselage.MASS, val=0)

        self.declare_partials(
            Aircraft.Fuselage.MASS,
            [
                Aircraft.Fuselage.MASS_COEFFICIENT,
                "fus_mass_full",
                Aircraft.Fuselage.WETTED_AREA,
                Aircraft.Fuselage.AVG_DIAMETER,
                Aircraft.TailBoom.LENGTH,
                "pylon_len",
                "min_dive_vel",
                Aircraft.Fuselage.PRESSURE_DIFFERENTIAL,
                Aircraft.Wing.ULTIMATE_LOAD_FACTOR,
                "MAT",
            ],
        )
        self.declare_partials(Aircraft.Design.STRUCTURE_MASS, "*")

    def compute(self, inputs, outputs):

        fus_wt_full = inputs["fus_mass_full"] * GRAV_ENGLISH_LBM
        c_fuselage = inputs[Aircraft.Fuselage.MASS_COEFFICIENT]
        fus_SA = inputs[Aircraft.Fuselage.WETTED_AREA]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        cabin_len_tailboom = inputs[Aircraft.TailBoom.LENGTH]
        pylon_len = inputs["pylon_len"]
        min_dive_vel = inputs["min_dive_vel"]
        p_diff_fus = inputs[Aircraft.Fuselage.PRESSURE_DIFFERENTIAL]
        ULF = inputs[Aircraft.Wing.ULTIMATE_LOAD_FACTOR]
        WAT = inputs["MAT"] * GRAV_ENGLISH_LBM
        CK8 = inputs[Aircraft.Wing.MASS_SCALER]
        total_wing_wt = inputs[Aircraft.Wing.MASS] * GRAV_ENGLISH_LBM
        CK9 = inputs[Aircraft.HorizontalTail.MASS_SCALER]
        htail_wt = inputs[Aircraft.HorizontalTail.MASS] * GRAV_ENGLISH_LBM
        CK10 = inputs[Aircraft.VerticalTail.MASS_SCALER]
        vtail_wt = inputs[Aircraft.VerticalTail.MASS] * GRAV_ENGLISH_LBM
        CK11 = inputs[Aircraft.Fuselage.MASS_SCALER]
        CK12 = inputs[Aircraft.LandingGear.TOTAL_MASS_SCALER]
        landing_gear_wt = inputs[Aircraft.LandingGear.TOTAL_MASS] * \
            GRAV_ENGLISH_LBM
        CK14 = inputs[Aircraft.Engine.POD_MASS_SCALER]
        sec_wt = inputs[Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS] * \
            GRAV_ENGLISH_LBM
        delta_struct_wt = inputs[Aircraft.Design.STRUCTURAL_MASS_INCREMENT] * \
            GRAV_ENGLISH_LBM

        fus_wt = (
            c_fuselage
            * (
                (fus_wt_full / 10000.0) ** 0.7
                * (fus_SA / 1000.0)
                * cabin_width
                * (cabin_len_tailboom + pylon_len) ** 0.5
                * np.log10(min_dive_vel)
                * (p_diff_fus + 1.0) ** 0.2
                * ULF**0.3
            )
            ** 0.508
            + WAT
        )

        outputs[Aircraft.Fuselage.MASS] = fus_wt / GRAV_ENGLISH_LBM
        outputs[Aircraft.Design.STRUCTURE_MASS] = (
            CK8 * total_wing_wt
            + CK9 * htail_wt
            + CK10 * vtail_wt
            + CK11 * fus_wt
            + CK12 * landing_gear_wt
            + CK14 * sec_wt
            + delta_struct_wt
        ) / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):

        fus_wt_full = inputs["fus_mass_full"] * GRAV_ENGLISH_LBM
        c_fuselage = inputs[Aircraft.Fuselage.MASS_COEFFICIENT]
        fus_SA = inputs[Aircraft.Fuselage.WETTED_AREA]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        cabin_len_tailboom = inputs[Aircraft.TailBoom.LENGTH]
        pylon_len = inputs["pylon_len"]
        min_dive_vel = inputs["min_dive_vel"]
        p_diff_fus = inputs[Aircraft.Fuselage.PRESSURE_DIFFERENTIAL]
        ULF = inputs[Aircraft.Wing.ULTIMATE_LOAD_FACTOR]
        WAT = inputs["MAT"] * GRAV_ENGLISH_LBM
        CK8 = inputs[Aircraft.Wing.MASS_SCALER]
        total_wing_wt = inputs[Aircraft.Wing.MASS] * GRAV_ENGLISH_LBM
        CK9 = inputs[Aircraft.HorizontalTail.MASS_SCALER]
        htail_wt = inputs[Aircraft.HorizontalTail.MASS] * GRAV_ENGLISH_LBM
        CK10 = inputs[Aircraft.VerticalTail.MASS_SCALER]
        vtail_wt = inputs[Aircraft.VerticalTail.MASS] * GRAV_ENGLISH_LBM
        CK11 = inputs[Aircraft.Fuselage.MASS_SCALER]
        CK12 = inputs[Aircraft.LandingGear.TOTAL_MASS_SCALER]
        landing_gear_wt = inputs[Aircraft.LandingGear.TOTAL_MASS] * \
            GRAV_ENGLISH_LBM
        CK14 = inputs[Aircraft.Engine.POD_MASS_SCALER]
        sec_wt = inputs[Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS] * \
            GRAV_ENGLISH_LBM
        delta_struct_wt = inputs[Aircraft.Design.STRUCTURAL_MASS_INCREMENT] * \
            GRAV_ENGLISH_LBM

        x = 0.508 * (
            (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * cabin_width
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        ) ** (0.508 - 1)
        fus_wt = (
            c_fuselage
            * (
                (fus_wt_full / 10000.0) ** 0.7
                * (fus_SA / 1000.0)
                * cabin_width
                * (cabin_len_tailboom + pylon_len) ** 0.5
                * np.log10(min_dive_vel)
                * (p_diff_fus + 1.0) ** 0.2
                * ULF**0.3
            )
            ** 0.508
            + WAT
        )
        dFusWt_dCFus = (
            (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * cabin_width
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        ) ** 0.508
        dFusWt_dFusWtFull = (
            c_fuselage
            * x
            * 0.7
            * (fus_wt_full / 10000.0) ** (0.7 - 1)
            * (1 / 10000)
            * (fus_SA / 1000.0)
            * cabin_width
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        )
        dFusWt_dFusSA = (
            c_fuselage
            * x
            * (fus_wt_full / 10000.0) ** 0.7
            * (1 / 1000.0)
            * cabin_width
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        )
        dFusWt_dCabWidth = (
            c_fuselage
            * x
            * (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        )
        dFusWt_dCabLenBoom = (
            c_fuselage
            * x
            * (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * cabin_width
            * 0.5
            * (cabin_len_tailboom + pylon_len) ** (-0.5)
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        )
        dFusWt_dPylonLen = (
            c_fuselage
            * x
            * (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * cabin_width
            * 0.5
            * (cabin_len_tailboom + pylon_len) ** (-0.5)
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        )
        dFusWt_dMinDiveVel = (
            c_fuselage
            * x
            * (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * cabin_width
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * 1
            / (min_dive_vel * np.log(10))
            * (p_diff_fus + 1.0) ** 0.2
            * ULF**0.3
        )
        dFusWt_dPdiffFus = (
            c_fuselage
            * x
            * (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * cabin_width
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * np.log10(min_dive_vel)
            * 0.2
            * (p_diff_fus + 1.0) ** (0.2 - 1)
            * ULF**0.3
        )
        dFusWt_dULF = (
            c_fuselage
            * x
            * (fus_wt_full / 10000.0) ** 0.7
            * (fus_SA / 1000.0)
            * cabin_width
            * (cabin_len_tailboom + pylon_len) ** 0.5
            * np.log10(min_dive_vel)
            * (p_diff_fus + 1.0) ** 0.2
            * 0.3
            * ULF ** (0.3 - 1)
        )

        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.MASS_COEFFICIENT] = \
            dFusWt_dCFus / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, "fus_mass_full"] = dFusWt_dFusWtFull
        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.WETTED_AREA] = \
            dFusWt_dFusSA / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.AVG_DIAMETER] = \
            dFusWt_dCabWidth / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, Aircraft.TailBoom.LENGTH] = \
            dFusWt_dCabLenBoom / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, "pylon_len"] = \
            dFusWt_dPylonLen / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, "min_dive_vel"] = \
            dFusWt_dMinDiveVel / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, Aircraft.Fuselage.PRESSURE_DIFFERENTIAL] = \
            dFusWt_dPdiffFus / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, Aircraft.Wing.ULTIMATE_LOAD_FACTOR] = \
            dFusWt_dULF / GRAV_ENGLISH_LBM
        J[Aircraft.Fuselage.MASS, "MAT"] = dFusWt_dWAT = 1

        J[Aircraft.Design.STRUCTURE_MASS,
            Aircraft.Fuselage.MASS_COEFFICIENT] = dFusWt_dCFus * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, "fus_mass_full"] = dFusWt_dFusWtFull * CK11
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.Fuselage.WETTED_AREA] = \
            dFusWt_dFusSA * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS,
            Aircraft.Fuselage.AVG_DIAMETER] = dFusWt_dCabWidth * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.TailBoom.LENGTH] = \
            dFusWt_dCabLenBoom * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, "pylon_len"] = \
            dFusWt_dPylonLen * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, "min_dive_vel"] = \
            dFusWt_dMinDiveVel * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS,
            Aircraft.Fuselage.PRESSURE_DIFFERENTIAL] = dFusWt_dPdiffFus * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS,
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR] = dFusWt_dULF * CK11 / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, "MAT"] = dFusWt_dWAT * CK11
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.Wing.MASS_SCALER] = \
            total_wing_wt / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.Wing.MASS] = CK8
        J[Aircraft.Design.STRUCTURE_MASS,
            Aircraft.HorizontalTail.MASS_SCALER] = htail_wt / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.HorizontalTail.MASS] = CK9
        J[Aircraft.Design.STRUCTURE_MASS,
            Aircraft.VerticalTail.MASS_SCALER] = vtail_wt / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.VerticalTail.MASS] = CK10
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.Fuselage.MASS_SCALER] = \
            fus_wt / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS,
            Aircraft.LandingGear.TOTAL_MASS_SCALER] = landing_gear_wt / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.LandingGear.TOTAL_MASS] = CK12
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.Engine.POD_MASS_SCALER] = \
            sec_wt / GRAV_ENGLISH_LBM
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS] = CK14
        J[Aircraft.Design.STRUCTURE_MASS, Aircraft.Design.STRUCTURAL_MASS_INCREMENT] = 1


class FuelMass(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Design.STRUCTURE_MASS, val=50461.0)
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS, val=1759,)
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400)
        self.add_input(
            "eng_comb_mass",
            val=14371.0,
            units="lbm",
            desc="WPSTAR: mass of dry engine and engine installation. Includes mass of electrical augmentation system.",
        )
        add_aviary_input(self, Aircraft.Controls.TOTAL_MASS, val=3895.0)
        add_aviary_input(self, Aircraft.Design.FIXED_EQUIPMENT_MASS, val=21089.0)
        add_aviary_input(self, Aircraft.Design.FIXED_USEFUL_LOAD, val=4932.0)
        self.add_input(
            "payload_mass_des", val=36000, units="lbm", desc="WPLDES: design payload"
        )
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1)
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041)
        add_aviary_input(self, Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        add_aviary_input(self, Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS, val=36000)

        self.add_input(
            "payload_mass_max",
            val=46040,
            units="lbm",
            desc="WPLMAX: maximum payload that the aircraft is being asked to carry (design payload + cargo)",
        )

        add_aviary_input(self, Aircraft.Fuel.FUEL_MARGIN, val=10)

        add_aviary_output(self, Mission.Design.FUEL_MASS, val=0)
        add_aviary_output(self, Aircraft.Propulsion.MASS, val=0)
        add_aviary_output(self, Mission.Design.FUEL_MASS_REQUIRED, val=0)
        self.add_output(
            "fuel_mass_min",
            val=0,
            units="lbm",
            desc="WFAMIN: minimum value of fuel mass (set when max payload is carried)",
        )

        self.declare_partials(
            Mission.Design.FUEL_MASS,
            [
                Mission.Design.GROSS_MASS,
                "eng_comb_mass",
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
                "payload_mass_des",
                Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER,
                Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT,
                Aircraft.Fuel.DENSITY,
                Aircraft.Fuel.FUEL_MARGIN,
            ],
        )
        self.declare_partials(Aircraft.Propulsion.MASS, [
                              "eng_comb_mass", Aircraft.Fuel.FUEL_SYSTEM_MASS], val=1)
        self.declare_partials(Mission.Design.FUEL_MASS_REQUIRED, [
                              Mission.Design.GROSS_MASS], val=1)
        self.declare_partials(
            Mission.Design.FUEL_MASS_REQUIRED,
            [
                "eng_comb_mass",
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
                Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS,
                Aircraft.Fuel.FUEL_SYSTEM_MASS,
            ],
            val=-1,
        )
        self.declare_partials("fuel_mass_min", [Mission.Design.GROSS_MASS], val=1)
        self.declare_partials(
            "fuel_mass_min",
            [
                "eng_comb_mass",
                Aircraft.Design.STRUCTURE_MASS,
                Aircraft.Controls.TOTAL_MASS,
                Aircraft.Design.FIXED_EQUIPMENT_MASS,
                Aircraft.Design.FIXED_USEFUL_LOAD,
                "payload_mass_max",
                Aircraft.Fuel.FUEL_SYSTEM_MASS,
            ],
            val=-1,
        )

    def compute(self, inputs, outputs):

        struct_wt = inputs[Aircraft.Design.STRUCTURE_MASS] * GRAV_ENGLISH_LBM
        fuel_sys_wt = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS] * GRAV_ENGLISH_LBM
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        eng_comb_wt = inputs["eng_comb_mass"] * GRAV_ENGLISH_LBM
        control_wt = inputs[Aircraft.Controls.TOTAL_MASS] * GRAV_ENGLISH_LBM
        fixed_equip_wt = inputs[Aircraft.Design.FIXED_EQUIPMENT_MASS] * \
            GRAV_ENGLISH_LBM
        useful_wt = inputs[Aircraft.Design.FIXED_USEFUL_LOAD] * \
            GRAV_ENGLISH_LBM
        payload_wt_des = inputs["payload_mass_des"] * GRAV_ENGLISH_LBM
        CK21 = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]
        c_mass_trend_fuel_sys = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT]
        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        payload_wt = inputs[Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS] * \
            GRAV_ENGLISH_LBM
        payload_wt_max = inputs["payload_mass_max"] * GRAV_ENGLISH_LBM
        fuel_margin = inputs[Aircraft.Fuel.FUEL_MARGIN]

        outputs[Mission.Design.FUEL_MASS] = (
            gross_wt_initial
            - eng_comb_wt
            - struct_wt
            - control_wt
            - fixed_equip_wt
            - useful_wt
            - payload_wt_des
        ) / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        ) / GRAV_ENGLISH_LBM
        outputs[Aircraft.Propulsion.MASS] = (
            eng_comb_wt + fuel_sys_wt) / GRAV_ENGLISH_LBM
        outputs[Mission.Design.FUEL_MASS_REQUIRED] = (
            gross_wt_initial
            - eng_comb_wt
            - struct_wt
            - control_wt
            - fixed_equip_wt
            - useful_wt
            - payload_wt
            - fuel_sys_wt
        ) / GRAV_ENGLISH_LBM
        outputs["fuel_mass_min"] = (
            gross_wt_initial
            - eng_comb_wt
            - struct_wt
            - control_wt
            - fixed_equip_wt
            - useful_wt
            - payload_wt_max
            - fuel_sys_wt
        ) / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):

        struct_wt = inputs[Aircraft.Design.STRUCTURE_MASS] * GRAV_ENGLISH_LBM
        fuel_sys_wt = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS] * GRAV_ENGLISH_LBM
        gross_wt_initial = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        eng_comb_wt = inputs["eng_comb_mass"] * GRAV_ENGLISH_LBM
        control_wt = inputs[Aircraft.Controls.TOTAL_MASS] * GRAV_ENGLISH_LBM
        fixed_equip_wt = inputs[Aircraft.Design.FIXED_EQUIPMENT_MASS] * \
            GRAV_ENGLISH_LBM
        useful_wt = inputs[Aircraft.Design.FIXED_USEFUL_LOAD] * \
            GRAV_ENGLISH_LBM
        payload_wt_des = inputs["payload_mass_des"] * GRAV_ENGLISH_LBM
        CK21 = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]
        c_mass_trend_fuel_sys = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT]
        rho_fuel = inputs[Aircraft.Fuel.DENSITY] * GRAV_ENGLISH_LBM
        payload_wt = inputs[Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS] * \
            GRAV_ENGLISH_LBM
        payload_wt_max = inputs["payload_mass_max"] * GRAV_ENGLISH_LBM
        fuel_margin = inputs[Aircraft.Fuel.FUEL_MARGIN]

        J[Mission.Design.FUEL_MASS, Mission.Design.GROSS_MASS] = 1 / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        )
        J[Mission.Design.FUEL_MASS, "eng_comb_mass"] = -1 / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Design.STRUCTURE_MASS] = -1 / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Controls.TOTAL_MASS] = -1 / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Design.FIXED_EQUIPMENT_MASS] = -1 / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Design.FIXED_USEFUL_LOAD] = -1 / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        )
        J[Mission.Design.FUEL_MASS, "payload_mass_des"] = -1 / (
            1.0
            + CK21 * c_mass_trend_fuel_sys * (1 + fuel_margin / 100) * 6.687 / rho_fuel
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER] = (
            -(
                gross_wt_initial
                - eng_comb_wt
                - struct_wt
                - control_wt
                - fixed_equip_wt
                - useful_wt
                - payload_wt_des
            )
            / (
                1.0
                + CK21
                * c_mass_trend_fuel_sys
                * (1 + fuel_margin / 100)
                * 6.687
                / rho_fuel
            )
            ** 2
            * c_mass_trend_fuel_sys
            * (1 + fuel_margin / 100)
            * 6.687
            / rho_fuel / GRAV_ENGLISH_LBM
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT] = (
            -(
                gross_wt_initial
                - eng_comb_wt
                - struct_wt
                - control_wt
                - fixed_equip_wt
                - useful_wt
                - payload_wt_des
            )
            / (
                1.0
                + CK21
                * c_mass_trend_fuel_sys
                * (1 + fuel_margin / 100)
                * 6.687
                / rho_fuel
            )
            ** 2
            * CK21
            * (1 + fuel_margin / 100)
            * 6.687
            / rho_fuel / GRAV_ENGLISH_LBM
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Fuel.DENSITY] = (
            -(
                gross_wt_initial
                - eng_comb_wt
                - struct_wt
                - control_wt
                - fixed_equip_wt
                - useful_wt
                - payload_wt_des
            )
            / (
                1.0
                + CK21
                * c_mass_trend_fuel_sys
                * (1 + fuel_margin / 100)
                * 6.687
                / rho_fuel
            )
            ** 2
            * (
                -CK21
                * c_mass_trend_fuel_sys
                * (1 + fuel_margin / 100)
                * 6.687
                / rho_fuel**2
            )
        )
        J[Mission.Design.FUEL_MASS, Aircraft.Fuel.FUEL_MARGIN] = (
            -(
                gross_wt_initial
                - eng_comb_wt
                - struct_wt
                - control_wt
                - fixed_equip_wt
                - useful_wt
                - payload_wt_des
            )
            / (
                1.0
                + CK21
                * c_mass_trend_fuel_sys
                * (1 + fuel_margin / 100)
                * 6.687
                / rho_fuel
            )
            ** 2
            * CK21
            * c_mass_trend_fuel_sys
            * (1 / 100)
            * 6.687
            / rho_fuel / GRAV_ENGLISH_LBM
        )


class FuelMassGroup(om.Group):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        # variables that are calculated at a higher level
        higher_level_inputs1 = ["wing_mounted_mass"]
        higher_level_inputs2 = [
            "min_dive_vel",
        ]
        higher_level_inputs3 = [
            "payload_mass_des",
            "payload_mass_max",
            "eng_comb_mass",
        ]

        # variables that are passed within the group but not used at a higher level
        connected_inputs1 = ["wingfuel_mass_min"]
        connected_inputs2 = [
            "fus_mass_full",
        ]
        connected_inputs5 = [
            "fuel_mass_min",
            "max_wingfuel_mass",
        ]

        connected_outputs1 = ["fus_mass_full"]
        connected_outputs3 = [
            "fuel_mass_min",
        ]
        connected_outputs4 = [
            "max_wingfuel_mass",
        ]
        connected_outputs5 = ["wingfuel_mass_min"]

        self.add_subsystem(
            "sys_and_full_fus",
            FuelSysAndFullFuselageMass(aviary_options=aviary_options),
            promotes_inputs=connected_inputs1
            + higher_level_inputs1
            + ["aircraft:*", "mission:*"],
            promotes_outputs=connected_outputs1 + ["aircraft:*"],
        )

        self.add_subsystem(
            "fus_and_struct",
            FuselageAndStructMass(aviary_options=aviary_options),
            promotes_inputs=connected_inputs2 + higher_level_inputs2 + ["aircraft:*"],
            promotes_outputs=["aircraft:*"],
        )

        self.add_subsystem(
            "fuel",
            FuelMass(aviary_options=aviary_options),
            promotes_inputs=higher_level_inputs3
            + ["aircraft:*", "mission:*"],
            promotes_outputs=connected_outputs3 + ["aircraft:*", "mission:*"],
        )

        self.add_subsystem(
            "fuel_and_oem",
            FuelAndOEMOutputs(aviary_options=aviary_options),
            promotes_inputs=["aircraft:*", "mission:*"],
            promotes_outputs=connected_outputs4 + ["aircraft:*"],
        )

        self.add_subsystem(
            "body_tank",
            BodyTankCalculations(aviary_options=aviary_options),
            promotes_inputs=connected_inputs5
            + ["aircraft:*", "mission:*"],
            promotes_outputs=connected_outputs5 + ["aircraft:*"],
        )

        self.set_input_defaults(Mission.Design.GROSS_MASS, val=175400, units="lbm")
        self.set_input_defaults(Aircraft.Wing.MASS, val=15830, units="lbm")
        self.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless")
        self.set_input_defaults(
            Aircraft.Fuel.FUEL_SYSTEM_MASS_COEFFICIENT, val=0.041, units="unitless")
        self.set_input_defaults(Aircraft.Fuel.DENSITY, val=6.687, units="lbm/galUS")
        self.set_input_defaults(Aircraft.Controls.TOTAL_MASS, val=3895, units="lbm")
        self.set_input_defaults(
            Aircraft.Design.FIXED_EQUIPMENT_MASS, val=21089.0, units="lbm")
        self.set_input_defaults(Aircraft.Design.FIXED_USEFUL_LOAD, val=4932, units="lbm")

        newton = self.nonlinear_solver = om.NewtonSolver()
        newton.options["atol"] = 1e-9
        newton.options["rtol"] = 1e-9
        newton.options["iprint"] = 2
        newton.options["maxiter"] = 10
        newton.options["solve_subsystems"] = True
        newton.options["max_sub_solves"] = 10
        newton.options["err_on_non_converge"] = True
        newton.options["reraise_child_analysiserror"] = False
        newton.linesearch = om.BoundsEnforceLS()
        newton.linesearch.options["bound_enforcement"] = "scalar"
        newton.linesearch.options["iprint"] = -1
        newton.options["err_on_non_converge"] = False

        self.linear_solver = om.DirectSolver(assemble_jac=True)



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import (
    distributed_engine_count_factor, distributed_thrust_factor)
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class TransportAirCondMass(om.ExplicitComponent):
    '''
    Calculates the mass of the air conditioning group using the transport/general
    aviation method. The methodology is based on the FLOPS weight equations,
    modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.AirConditioning.MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Avionics.MASS, val=0.0)

        add_aviary_input(self, Aircraft.Fuselage.MAX_HEIGHT, val=0.0)

        add_aviary_input(self, Aircraft.Fuselage.PLANFORM_AREA, val=0.0)

        add_aviary_output(self, Aircraft.AirConditioning.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        pax = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')

        scaler = inputs[Aircraft.AirConditioning.MASS_SCALER]
        avionics_wt = inputs[Aircraft.Avionics.MASS] * GRAV_ENGLISH_LBM
        height = inputs[Aircraft.Fuselage.MAX_HEIGHT]
        planform = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        outputs[Aircraft.AirConditioning.MASS] = \
            ((3.2 * (planform * height)**0.6 + 9 * pax**0.83)
             * max_mach + 0.075 * avionics_wt) * scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        pax = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')

        scaler = inputs[Aircraft.AirConditioning.MASS_SCALER]
        avionics_wt = inputs[Aircraft.Avionics.MASS] * GRAV_ENGLISH_LBM
        height = inputs[Aircraft.Fuselage.MAX_HEIGHT]
        planform = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        planform_exp = planform**0.6
        height_exp = height**0.6
        pax_exp = pax**0.83

        J[Aircraft.AirConditioning.MASS, Aircraft.AirConditioning.MASS_SCALER] = (
            (3.2 * planform_exp * height_exp + 9 * pax_exp) *
            max_mach + 0.075 * avionics_wt) / GRAV_ENGLISH_LBM

        J[Aircraft.AirConditioning.MASS, Aircraft.Avionics.MASS] = \
            0.075 * scaler

        J[Aircraft.AirConditioning.MASS, Aircraft.Fuselage.MAX_HEIGHT] = \
            1.92 * planform_exp * height**-0.4 * max_mach * scaler / GRAV_ENGLISH_LBM

        J[Aircraft.AirConditioning.MASS, Aircraft.Fuselage.PLANFORM_AREA] = \
            1.92 * planform**-0.4 * height_exp * max_mach * scaler / GRAV_ENGLISH_LBM


class AltAirCondMass(om.ExplicitComponent):
    '''
    Calculates the mass of the air conditioning group using the alternate method.
    The methodology is based on the FLOPS weight equations, modified to output
    mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.AirConditioning.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.AirConditioning.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials(of=Aircraft.AirConditioning.MASS, wrt='*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_pax = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')

        scaler = inputs[Aircraft.AirConditioning.MASS_SCALER]

        outputs[Aircraft.AirConditioning.MASS] = \
            26.0 * num_pax * scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_pax = aviary_options.get_val(
            Aircraft.CrewPayload.NUM_PASSENGERS, units='unitless')

        J[Aircraft.AirConditioning.MASS, Aircraft.AirConditioning.MASS_SCALER] = \
            26.0 * num_pax / GRAV_ENGLISH_LBM



"""
This is a slightly more complex Aviary example of running a coupled aircraft design-mission optimization.
It runs the same mission as the `run_basic_aviary_example.py` script, but it uses the AviaryProblem class to set up the problem.
This exposes more options and flexibility to the user and uses the "Level 2" API within Aviary.

We define a `phase_info` object, which tells Aviary how to model the mission.
Here we have climb, cruise, and descent phases.
We then call the correct methods in order to set up and run an Aviary optimization problem.
This performs a coupled design-mission optimization and outputs the results from Aviary into the `reports` folder.
"""
import aviary.api as av


phase_info = {
    "pre_mission": {"include_takeoff": False, "optimize_mass": True},
    "climb": {
        "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
        "user_options": {
            "optimize_mach": False,
            "optimize_altitude": False,
            "polynomial_control_order": 1,
            "num_segments": 5,
            "order": 3,
            "solve_for_distance": False,
            "initial_mach": (0.2, "unitless"),
            "final_mach": (0.72, "unitless"),
            "mach_bounds": ((0.18, 0.74), "unitless"),
            "initial_altitude": (0.0, "ft"),
            "final_altitude": (32000.0, "ft"),
            "altitude_bounds": ((0.0, 34000.0), "ft"),
            "throttle_enforcement": "path_constraint",
            "fix_initial": True,
            "constrain_final": False,
            "fix_duration": False,
            "initial_bounds": ((0.0, 0.0), "min"),
            "duration_bounds": ((64.0, 192.0), "min"),
        },
        "initial_guesses": {"times": ([0, 128], "min")},
    },
    "cruise": {
        "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
        "user_options": {
            "optimize_mach": False,
            "optimize_altitude": False,
            "polynomial_control_order": 1,
            "num_segments": 5,
            "order": 3,
            "solve_for_distance": False,
            "initial_mach": (0.72, "unitless"),
            "final_mach": (0.72, "unitless"),
            "mach_bounds": ((0.7, 0.74), "unitless"),
            "initial_altitude": (32000.0, "ft"),
            "final_altitude": (34000.0, "ft"),
            "altitude_bounds": ((23000.0, 38000.0), "ft"),
            "throttle_enforcement": "boundary_constraint",
            "fix_initial": False,
            "constrain_final": False,
            "fix_duration": False,
            "initial_bounds": ((64.0, 192.0), "min"),
            "duration_bounds": ((56.5, 169.5), "min"),
        },
        "initial_guesses": {"times": ([128, 113], "min")},
    },
    "descent": {
        "subsystem_options": {"core_aerodynamics": {"method": "computed"}},
        "user_options": {
            "optimize_mach": False,
            "optimize_altitude": False,
            "polynomial_control_order": 1,
            "num_segments": 5,
            "order": 3,
            "solve_for_distance": False,
            "initial_mach": (0.72, "unitless"),
            "final_mach": (0.36, "unitless"),
            "mach_bounds": ((0.34, 0.74), "unitless"),
            "initial_altitude": (34000.0, "ft"),
            "final_altitude": (500.0, "ft"),
            "altitude_bounds": ((0.0, 38000.0), "ft"),
            "throttle_enforcement": "path_constraint",
            "fix_initial": False,
            "constrain_final": True,
            "fix_duration": False,
            "initial_bounds": ((120.5, 361.5), "min"),
            "duration_bounds": ((29.0, 87.0), "min"),
        },
        "initial_guesses": {"times": ([241, 58], "min")},
    },
    "post_mission": {
        "include_landing": False,
        "constrain_range": True,
        "target_range": (1906, "nmi"),
    },
}


prob = av.AviaryProblem()

# Load aircraft and options data from user
# Allow for user overrides here
prob.load_inputs('models/test_aircraft/aircraft_for_bench_FwFm.csv', phase_info)


# Preprocess inputs
prob.check_and_preprocess_inputs()

prob.add_pre_mission_systems()

prob.add_phases()

prob.add_post_mission_systems()

# Link phases and variables
prob.link_phases()

prob.add_driver("SLSQP", max_iter=100)

prob.add_design_variables()

# Load optimization problem formulation
# Detail which variables the optimizer can control
prob.add_objective()

prob.setup()

prob.set_initial_guesses()

prob.run_aviary_problem()



import unittest

import aviary.api as av
from aviary.utils.compare_hierarchies import compare_hierarchies_to_merge

AviaryAircraft = av.Aircraft
AviaryMission = av.Mission


class Aircraft1(AviaryAircraft):
    ''' Simple variable hierarchy extension for testing. '''

    CG = "aircraft:center_of_gravity"

    class LandingGear(AviaryAircraft.LandingGear):
        SHAPE = 'aircraft:landing_gear:shape'

    class Wing(AviaryAircraft.Wing):
        CHARACTER = 'aircraft:wing:character'

    class Jury:
        MASS = 'aircraft:jury:mass'

    class Engine(AviaryAircraft.Engine):
        COLOR = 'aircraft:engine:color'

    class Fuselage(AviaryAircraft.Fuselage):

        class Nose:
            ECCENTRICITY = 'aircraft:fuselage:nose:eccentricity'


class Aircraft2(AviaryAircraft):
    ''' Simple variable hierarchy extension for testing. '''

    MASS = "aircraft:mass"
    CG = 'aircraft:center_of_gravity'

    class HorizontalTail(AviaryAircraft.HorizontalTail):
        # add variable that is in the core hierarchy and give it the right value
        AREA = 'aircraft:horizontal_tail:area'
        MEAN_AERO_CHORD = "aircraft:horizontal_tail:mean_aerodynamic_chord"

        class Elevator:
            AREA = "aircraft:horizontal_tail:elevator_area_dist"

    class Wing(AviaryAircraft.Wing):

        AERO_CENTER = "aircraft:wing:aerodynamic_center"
        CHORD = "aircraft:wing:chord"
        AIRFOIL_TECHNOLOGY = 'aircraft:wing:airfoil_technology'

        class Flap:
            AREA = "aircraft:wing:flap_area_dist"
            ROOT_CHORD = "aircraft:wing:flap_root_chord_dist"
            SPAN = "aircraft:wing:flap_span_dist"

    class Jury:
        MASS = 'aircraft:jury:mass'

    class Engine(AviaryAircraft.Engine):
        COLOR = 'aircraft:engine:color'

    class Fuselage(AviaryAircraft.Fuselage):

        class Nose:
            ECCENTRICITY = 'aircraft:fuselage:nose:eccentricity'


class Aircraft3(AviaryAircraft):
    ''' Simple variable hierarchy extension for testing. '''

    class HorizontalTail:
        AREA = 'the_wrong_value'  # add variable that is in the core hierarchy and give it the wrong value


class Aircraft_a:
    ''' Simple variable hierarchy for testing. '''

    A1 = 'a1'

    class Subclass:
        A2 = 'a2'

        class Subsubclass:
            A3 = 'a3'

            class Subsubsubclass:
                A4 = 'a4'

                class Subsubsubsubclass:
                    A5 = 'a5'

                    class Sub5:
                        A6 = 'a6'

                        class Sub6:
                            A7 = 'a7'

                            class Sub7:
                                A8 = 'a8'

                                class Sub8:
                                    A9 = 'a9'

                                    class Sub9:
                                        A10 = 'a10'

                                        class Sub10:
                                            A11 = 'a11'


class Aircraft_b:
    ''' Simple variable hierarchy for testing. '''

    A1 = 'a1'

    class Subclass:
        A2 = 'a2'

        class Subsubclass:
            A3 = 'a3'

            class Subsubsubclass:
                A4 = 'a4'

                class Subsubsubsubclass:
                    A5 = 'a5'

                    class Sub5:
                        A6 = 'a6'

                        class Sub6:
                            A7 = 'a7'

                            class Sub7:
                                A8 = 'a8'

                                class Sub8:
                                    A9 = 'a9'

                                    class Sub9:
                                        A10 = 'a10'

                                        class Sub10:
                                            A11 = 'a11'


class Aircraft_c:
    ''' Simple variable hierarchy for testing. '''

    A1 = 'a1'

    class Subclass:
        A2 = 'a2'

        class Subsubclass:
            A3 = 'a3'

            class Subsubsubclass:
                A4 = 'a4'

                class Subsubsubsubclass:
                    A5 = 'a5'

                    class Sub5:
                        A6 = 'a6'

                        class Sub6:
                            A7 = 'a7'

                            class Sub7:
                                A8 = 'a8'

                                class Sub8:
                                    A9 = 'a9'

                                    class Sub9:
                                        A10 = 'a10'

                                        class Sub10:
                                            A11 = 'a1'


class CompareHierarchiesTest(unittest.TestCase):
    def test_compare_successful(self):
        # this shouldn't throw an error
        compare_hierarchies_to_merge([Aircraft1, Aircraft2, Aircraft_a, Aircraft_b])

    def test_compare_error_in_extended(self):
        self.maxDiff = None
        extension_error_msg = "You have attempted to merge two variable hierarchies together that have the same variable with a different string name associated to it. The offending variable is 'A11'. In 'Aircraft_a.Subclass.Subsubclass.Subsubsubclass.Subsubsubsubclass.Sub5.Sub6.Sub7.Sub8.Sub9.Sub10' it has a value of 'a11' and in 'Aircraft_c.Subclass.Subsubclass.Subsubsubclass.Subsubsubsubclass.Sub5.Sub6.Sub7.Sub8.Sub9.Sub10' it has a value of 'a1'."
        with self.assertRaises(ValueError) as cm:
            compare_hierarchies_to_merge([Aircraft_a, Aircraft_c])
        self.assertEqual(str(cm.exception), extension_error_msg)

    def test_compare_error_in_updated(self):
        self.maxDiff = None
        updated_error_msg = "You have attempted to merge two variable hierarchies together that have the same variable with a different string name associated to it. The offending variable is 'AREA'. In 'Aircraft3.HorizontalTail' it has a value of 'the_wrong_value' and in 'Aircraft.HorizontalTail' it has a value of 'aircraft:horizontal_tail:area'."
        with self.assertRaises(ValueError) as cm:
            compare_hierarchies_to_merge([Aircraft3, Aircraft_c])
        self.assertEqual(str(cm.exception), updated_error_msg)


if __name__ == '__main__':
    unittest.main()



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from parameterized import parameterized

from aviary.subsystems.aerodynamics.flops_based.lift import (LiftEqualsWeight,
                                                             SimpleLift)
from aviary.utils.aviary_values import AviaryValues
from aviary.validation_cases.validation_tests import (get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Dynamic

data_sets = get_flops_case_names(
    only=['LargeSingleAisle1FLOPS', 'LargeSingleAisle2FLOPS', 'N3CC'])


class SimpleLiftTest(unittest.TestCase):

    @parameterized.expand(data_sets, name_func=print_case)
    def test_case(self, case_name):
        flops_inputs = get_flops_inputs(case_name)

        dynamics_data: AviaryValues = dynamics_test_data[case_name]

        # area = 4 digits precision
        inputs_keys = (Aircraft.Wing.AREA,)

        # dynamic pressure = 4 digits precision
        # lift coefficient = 5 digits precision
        dynamics_keys = (Dynamic.Mission.DYNAMIC_PRESSURE, 'cl')

        # lift = 6 digits precision
        outputs_keys = (Dynamic.Mission.LIFT,)

        # use lowest precision from all available data
        tol = 1e-4

        prob = om.Problem()
        model = prob.model

        q, _ = dynamics_data.get_item(Dynamic.Mission.DYNAMIC_PRESSURE)
        nn = len(q)
        model.add_subsystem('simple_lift', SimpleLift(num_nodes=nn), promotes=['*'])

        prob.setup(force_alloc_complex=True)

        for key in inputs_keys:
            val, units = flops_inputs.get_item(key)
            prob.set_val(key, val, units)

        for key in dynamics_keys:
            val, units = dynamics_data.get_item(key)
            prob.set_val(key, val, units)

        prob.run_model()

        for key in outputs_keys:
            try:
                desired, units = dynamics_data.get_item(key)
                actual = prob.get_val(key, units)

                assert_near_equal(actual, desired, tol)

            except ValueError as error:
                msg = f'"{key}": {error!s}'

                raise ValueError(msg) from None

        data = prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(data, atol=2.5e-10, rtol=1e-12)


class LiftEqualsWeightTest(unittest.TestCase):

    @parameterized.expand(data_sets, name_func=print_case)
    def test_case(self, case_name):
        flops_inputs = get_flops_inputs(case_name)

        dynamics_data: AviaryValues = dynamics_test_data[case_name]

        # area = 4 digits precision
        inputs_keys = (Aircraft.Wing.AREA,)

        # dynamic pressure = 4 digits precision
        # mass = 6 digits precision
        dynamics_keys = (Dynamic.Mission.DYNAMIC_PRESSURE, Dynamic.Mission.MASS)

        # lift coefficient = 5 digits precision
        # lift = 6 digits precision
        outputs_keys = ('cl', Dynamic.Mission.LIFT)

        # use lowest precision from all available data
        tol = 1e-4

        prob = om.Problem()
        model = prob.model

        q, _ = dynamics_data.get_item(Dynamic.Mission.DYNAMIC_PRESSURE)
        nn = len(q)

        model.add_subsystem(
            'lift_equals_weight', LiftEqualsWeight(num_nodes=nn), promotes=['*'])

        prob.setup(force_alloc_complex=True)

        for key in inputs_keys:
            val, units = flops_inputs.get_item(key)
            prob.set_val(key, val, units)

        for key in dynamics_keys:
            val, units = dynamics_data.get_item(key)
            prob.set_val(key, val, units)

        prob.run_model()

        for key in outputs_keys:
            try:
                desired, units = dynamics_data.get_item(key)
                actual = prob.get_val(key, units)

                assert_near_equal(actual, desired, tol)

            except ValueError as error:
                msg = f'"{key}": {error!s}'

                raise ValueError(msg) from None

        data = prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(data, atol=2.5e-10, rtol=1e-12)


# dynamic test data taken from the baseline FLOPS output for each case
#    - first: # DETAILED FLIGHT SEGMENT SUMMARY
#        - first: SEGMENT... CRUISE
#            - first three points
dynamics_test_data = {}

dynamics_test_data['LargeSingleAisle1FLOPS'] = _dynamics_data = AviaryValues()
_dynamics_data.set_val(Dynamic.Mission.DYNAMIC_PRESSURE, [
                       206.0, 205.6, 205.4], 'lbf/ft**2')
_dynamics_data.set_val('cl', [0.62630, 0.62623, 0.62619])
_dynamics_data.set_val(Dynamic.Mission.LIFT, [176751., 176400., 176185.], 'lbf')
_dynamics_data.set_val(Dynamic.Mission.MASS, [176751., 176400., 176185.], 'lbm')

dynamics_test_data['LargeSingleAisle2FLOPS'] = _dynamics_data = AviaryValues()
_dynamics_data.set_val(Dynamic.Mission.DYNAMIC_PRESSURE, [
                       215.4, 215.4, 215.4], 'lbf/ft**2')
_dynamics_data.set_val('cl', [0.58761, 0.58578, 0.57954])
_dynamics_data.set_val(Dynamic.Mission.LIFT, [169730., 169200., 167400.], 'lbf')
_dynamics_data.set_val(Dynamic.Mission.MASS, [169730., 169200., 167400.], 'lbm')

dynamics_test_data['N3CC'] = _dynamics_data = AviaryValues()
_dynamics_data.set_val(Dynamic.Mission.DYNAMIC_PRESSURE, [
                       208.4, 288.5, 364.0], 'lbf/ft**2')
_dynamics_data.set_val('cl', [0.50651, 0.36573, 0.28970])
_dynamics_data.set_val(Dynamic.Mission.LIFT, [128777., 128721., 128667.], 'lbf')
_dynamics_data.set_val(Dynamic.Mission.MASS, [128777., 128721., 128667.], 'lbm')


if __name__ == "__main__":
    unittest.main()
